# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl
import math


def can_use_int32_index(tensor):
    # This function checks if the tensor can use int32 indices
    return tensor.numel() < 2**31

# Kernel 1: argmax_kernel_1
@triton.jit
def argmax_kernel_1(
    inp,
    mid_value,
    mid_index,
    M,
    BLOCK_SIZE: tl.constexpr,
    INT64_INDEX: tl.constexpr = False,
):
    pid = tl.program_id(0)
    if INT64_INDEX:
        pid = pid.to(tl.int64)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < M
    inp_val = tl.load(inp_ptrs, mask=mask, other=-float("inf"))
    max_val, max_index = tl.max(inp_val, axis=0, return_indices=True)
    max_index = max_index + pid * BLOCK_SIZE
    mid_value_ptr = mid_value + pid
    max_index_ptr = mid_index + pid
    tl.store(mid_value_ptr, max_val)
    tl.store(max_index_ptr, max_index)

# Kernel 2: argmax_kernel_2
@triton.jit
def argmax_kernel_2(mid_value, mid_index, out, mid_size, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid_value + offset
    mask = offset < mid_size
    mid_val = tl.load(mid_ptrs, mask=mask, other=-float("inf"))
    index_val = tl.argmax(mid_val, axis=0)
    mid_index_ptrs = mid_index + index_val
    out_val = tl.load(mid_index_ptrs)
    tl.store(out, out_val)

# Kernel 3: argmax_kernel
@triton.jit
def argmax_kernel(
    inp,
    out_index,
    M,
    N,
    K,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    INT64_INDEX: tl.constexpr = False,
):
    pid_m = tl.program_id(0)
    pid_k = tl.program_id(1)
    if INT64_INDEX:
        pid_m = pid_m.to(tl.int64)
        pid_k = pid_k.to(tl.int64)
    m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)

    max_values = tl.full([BLOCK_M], dtype=tl.float32, value=float("-inf"))
    argmax_values = tl.full([BLOCK_M], dtype=tl.int64, value=0)
    for start_n in range(0, N, BLOCK_N):
        n_offset = start_n + tl.arange(0, BLOCK_N)
        offset = m_offset[:, None] * N * K + n_offset[None, :] * K + pid_k
        mask = m_offset[:, None] < M and n_offset[None, :] < N
        inp_ptrs = inp + offset
        inp_vals = tl.load(inp_ptrs, mask=mask, other=-float("inf"))
        local_max, local_argmax = tl.max(
            inp_vals, 1, return_indices=True, return_indices_tie_break_left=True
        )
        update = local_max > max_values
        max_values = tl.where(update, local_max, max_values)
        argmax_values = tl.where(update, start_n + local_argmax, argmax_values)

    offset_index = m_offset * K + pid_k
    out_index_ptrs = out_index + offset_index
    mask1 = m_offset < M
    tl.store(out_index_ptrs, argmax_values, mask=mask1)

# Function calling the kernels
def argmax(inp, dim=None, keepdim=False, *, dtype=None):
    if dim is None:
        M = inp.numel()
        if dtype is None:
            dtype = inp.dtype
        block_size = triton.next_power_of_2(math.ceil(math.sqrt(M)))
        mid_size = triton.cdiv(M, block_size)
        block_mid = triton.next_power_of_2(mid_size)
        use_int64_index = not can_use_int32_index(inp)

        mid_value = torch.empty((mid_size,), dtype=dtype, device=inp.device)
        mid_index = torch.empty((mid_size,), dtype=torch.int64, device=inp.device)
        if keepdim:
            shape = list(inp.shape)
            for i in range(0, inp.dim()):
                shape[i] = 1
            out = torch.empty(shape, dtype=torch.int64, device=inp.device)
        else:
            out = torch.empty([], dtype=torch.int64, device=inp.device)

        with torch.cuda.device(inp.device):
            argmax_kernel_1[(mid_size, 1, 1)](
                inp,
                mid_value,
                mid_index,
                M,
                block_size,
                INT64_INDEX=use_int64_index,
            )
            argmax_kernel_2[(1, 1, 1)](mid_value, mid_index, out, mid_size, block_mid)
        return out
    else:
        assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
        shape = inp.shape
        dim = dim % inp.ndim
        N = shape[dim]
        M = math.prod(shape[:dim])
        K = inp.numel() // M // N

        inp = inp.contiguous()
        use_int64_index = not can_use_int32_index(inp)

        shape_list = list(shape)
        shape_list[dim] = 1
        out_index = torch.empty(shape_list, dtype=torch.int64, device=inp.device)
        if not keepdim:
            out_index = torch.squeeze(out_index, dim)

        BLOCK_M = 128  # Example, adjust as needed
        BLOCK_N = 128  # Example, adjust as needed

        grid = lambda meta: (
            triton.cdiv(M, meta["BLOCK_M"]),
            K,
        )
        with torch.cuda.device(inp.device):
            argmax_kernel[grid](
                inp,
                out_index,
                M,
                N,
                K,
                BLOCK_M=BLOCK_M,
                BLOCK_N=BLOCK_N,
                INT64_INDEX=use_int64_index,
            )

        return out_index



##################################################################################################################################################


import torch

def test_argmax():
    results = {}

    # Test case 1: 1D input tensor
    inp = torch.randn(1024, device='cuda')
    results['test_case_1'] = argmax(inp)

    # Test case 2: 2D input tensor, dim=0
    inp = torch.randn(1024, 1024, device='cuda')
    results['test_case_2'] = argmax(inp, dim=0)

    # Test case 3: 2D input tensor, dim=1
    inp = torch.randn(1024, 1024, device='cuda')
    results['test_case_3'] = argmax(inp, dim=1)

    # Test case 4: 3D input tensor
    inp = torch.randn(64, 128, 256, device='cuda')
    results['test_case_4'] = argmax(inp, dim=2)

    return results

# Run the test
result_gold = test_argmax()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl
import torch

@triton.jit
def matmul_kernel(c_ptr, a_ptr, b_ptr, BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr):
    pid_m = tl.program_id(axis=0)
    pid_n = tl.program_id(axis=1)

    M, N, K = 4096, 4096, 4096
    stride_am = 4096
    stride_ak = 1
    stride_bk = 4096
    stride_bn = 1
    stride_cm = 4096
    stride_cn = 1

    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)

    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)

    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        a = tl.load(a_ptrs)
        b = tl.load(b_ptrs)
        accumulator = tl.dot(a, b, accumulator)
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk

    c = tl.cast(accumulator, tl.float16)
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + stride_cm * offs_cm[:, None] + stride_cn * offs_cn[None, :]
    tl.store(c_ptrs, c)

def matmul(c, a, b, M, N, K, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K):
    matmul_kernel[triton.cdiv(M, BLOCK_SIZE_M), triton.cdiv(N, BLOCK_SIZE_N)](
        c, a, b, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K
    )



##################################################################################################################################################


def test_matmul():
    BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K = 64, 128, 64
    M, N, K = 4096, 4096, 4096

    # Initialize matrices on CUDA device
    c = torch.empty((M, N), device='cuda:0', dtype=torch.float16)
    a = torch.rand((M, K), device='cuda:0', dtype=torch.float16)
    b = torch.rand((K, N), device='cuda:0', dtype=torch.float16)

    # Call the matmul function multiple times
    test_case_1 = matmul(c, a, b, M, N, K, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K)

    # Additional test cases to cover more branches
    BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K = 128, 64, 128
    test_case_2 = matmul(c, a, b, M, N, K, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K)

    BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K = 256, 256, 64
    test_case_3 = matmul(c, a, b, M, N, K, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K)

    BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K = 32, 32, 32
    test_case_4 = matmul(c, a, b, M, N, K, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K)

    return {
        "test_case_1": test_case_1,
        "test_case_2": test_case_2,
        "test_case_3": test_case_3,
        "test_case_4": test_case_4
    }

result_gold = test_matmul()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import math
import torch
import triton
import triton.language as tl

_BLOCK_N = 64
_BLOCK_M = 64

@triton.jit
def _attn_fwd_inner(acc, l_i, m_i, q,
                    K_block_ptr, V_block_ptr,
                    start_m, qk_scale, N_CTX,
                    sliding_window_offset, sliding_window_size,
                    BLOCK_M: tl.constexpr, BLOCK_DMODEL: tl.constexpr, BLOCK_N: tl.constexpr, SLIDING_WINDOW: tl.constexpr,
                    IS_EVEN_M: tl.constexpr, IS_EVEN_N: tl.constexpr, COMPLEMENT_SLIDING_WINDOW: tl.constexpr
                ):
    # range of values handled by this stage
    if SLIDING_WINDOW and not COMPLEMENT_SLIDING_WINDOW:
        if COMPLEMENT_SLIDING_WINDOW:
            lo = 0
            hi = (((start_m + 1) * BLOCK_M + sliding_window_offset - sliding_window_size + BLOCK_N - 1) // BLOCK_N) * BLOCK_N
        else:
            lo = ((start_m * BLOCK_M + sliding_window_offset - sliding_window_size + 1) // BLOCK_N) * BLOCK_N
            hi = ((((start_m + 1) * BLOCK_M - 1) + sliding_window_offset + BLOCK_N) // BLOCK_N) * BLOCK_N
            if lo < 0:
                lo = 0
            if hi > N_CTX:
                hi = N_CTX

            lo = tl.multiple_of(lo, BLOCK_N)
            K_block_ptr = tl.advance(K_block_ptr, (0, lo))
            V_block_ptr = tl.advance(V_block_ptr, (lo, 0))
    else:
        lo, hi = 0, N_CTX

    # loop over k, v and update accumulator
    for start_n in range(lo, hi, BLOCK_N):
        start_n = tl.multiple_of(start_n, BLOCK_N)
        # -- compute qk ----
        if IS_EVEN_N:
            k = tl.load(K_block_ptr)
        else:
            k = tl.load(K_block_ptr, boundary_check=(0, 1), padding_option="zero")

        qk = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
        qk += tl.dot(q, k)
        qk = qk * qk_scale

        if SLIDING_WINDOW:
            dist = tl.arange(0, BLOCK_M)[:, None] - tl.arange(0, BLOCK_N)[None, :] \
                   + start_m * BLOCK_M - start_n + sliding_window_offset

            if COMPLEMENT_SLIDING_WINDOW:
                mask = (dist >= sliding_window_size)
            else:
                mask = (dist >= 0) & (dist < sliding_window_size)

            qk = tl.where(mask, qk, float("-inf"))

        if not IS_EVEN_N:
            qk = tl.where(((tl.arange(0, BLOCK_N) + start_n) < N_CTX)[None, :], qk, float("-inf"))

        m_ij = tl.maximum(m_i, tl.max(qk, 1))
        qk = qk - m_ij[:, None]
        p = tl.math.exp2(qk)

        if SLIDING_WINDOW:
            p = tl.where(mask, p, 0)

        if not IS_EVEN_N:
            p = tl.where(((tl.arange(0, BLOCK_N) + start_n) < N_CTX)[None, :], p, 0)

        l_ij = tl.sum(p, 1)
        # -- update m_i and l_i
        tmp = m_i - m_ij
        alpha_mask = (tmp != tmp)  # check nan
        alpha = tl.math.exp2(tmp)
        alpha = tl.where(alpha_mask, 1., alpha)
        l_i = l_i * alpha + l_ij
        # -- update output accumulator --
        acc = acc * alpha[:, None]
        # update acc
        if IS_EVEN_N:
            v = tl.load(V_block_ptr)
        else:
            v = tl.load(V_block_ptr, boundary_check=(0, 1), padding_option="zero")

        acc += tl.dot(p.to(v.dtype), v)
        # update m_i and l_i
        m_i = m_ij
        V_block_ptr = tl.advance(V_block_ptr, (BLOCK_N, 0))
        K_block_ptr = tl.advance(K_block_ptr, (0, BLOCK_N))

    return acc, l_i, m_i


@triton.heuristics(
    {
        "IS_EVEN_M": lambda args: args["N_CTX"] % args["BLOCK_M"] == 0,
        "IS_EVEN_N": lambda args: args["NKV_CTX"] % args["BLOCK_N"] == 0,
    }
)
@triton.jit
def _attn_fwd(Q, K, V, sm_scale, M, Out, L,#
              stride_qz, stride_qh, stride_qm, stride_qk,  #
              stride_kz, stride_kh, stride_kn, stride_kk,  #
              stride_vz, stride_vh, stride_vk, stride_vn,  #
              stride_oz, stride_oh, stride_om, stride_on,  #
              Z, H, H_KV, #
              N_CTX,  #
              ROUND_CTX,
              NKV_CTX,
              sliding_window_offset,
              sliding_window_size,
              IS_EVEN_M: tl.constexpr,
              IS_EVEN_N: tl.constexpr,
              BLOCK_M: tl.constexpr,  #
              BLOCK_DMODEL: tl.constexpr,  #
              BLOCK_N: tl.constexpr,  #
              END: tl.constexpr,
              INIT: tl.constexpr,
              SLIDING_WINDOW: tl.constexpr,
              COMPLEMENT_SLIDING_WINDOW: tl.constexpr
            ):

    start_m = tl.program_id(0)
    off_hz = tl.program_id(1)
    off_z = off_hz // H
    off_h = off_hz % H
    off_hkv = off_h // (H//H_KV)
    q_offset = off_z.to(tl.int64) * stride_qz + off_h.to(tl.int64) * stride_qh
    k_offset = off_z.to(tl.int64) * stride_kz + off_hkv.to(tl.int64) * stride_kh
    v_offset = off_z.to(tl.int64) * stride_vz + off_hkv.to(tl.int64) * stride_vh
    o_offset = off_z.to(tl.int64) * stride_oz + off_h.to(tl.int64) * stride_oh

    # block pointers
    Q_block_ptr = tl.make_block_ptr(
        base=Q + q_offset,
        shape=(N_CTX, BLOCK_DMODEL),
        strides=(stride_qm, stride_qk),
        offsets=(start_m * BLOCK_M, 0),
        block_shape=(BLOCK_M, BLOCK_DMODEL),
        order=(1, 0),
    )
    V_block_ptr = tl.make_block_ptr(
        base=V + v_offset,
        shape=(NKV_CTX, BLOCK_DMODEL),
        strides=(stride_vk, stride_vn),
        offsets=(0, 0),
        block_shape=(BLOCK_N, BLOCK_DMODEL),
        order=(1, 0),
    )
    K_block_ptr = tl.make_block_ptr(
        base=K + k_offset,
        shape=(BLOCK_DMODEL, NKV_CTX),
        strides=(stride_kk, stride_kn),
        offsets=(0, 0),
        block_shape=(BLOCK_DMODEL, BLOCK_N),
        order=(0, 1),
    )
    O_block_ptr = tl.make_block_ptr(
        base=Out + o_offset,
        shape=(ROUND_CTX, BLOCK_DMODEL),
        strides=(stride_om, stride_on),
        offsets=(start_m * BLOCK_M, 0),
        block_shape=(BLOCK_M, BLOCK_DMODEL),
        order=(1, 0),
    )
    # initialize offsets
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    # initialize pointer to m and l
    m_ptrs = M + off_hz * ROUND_CTX + offs_m
    l_ptrs = L + off_hz * ROUND_CTX + offs_m
    if INIT:
        m_i = tl.zeros([BLOCK_M], dtype=tl.float32) - float("inf")
        l_i = tl.zeros([BLOCK_M], dtype=tl.float32) + 1.0
        acc = tl.zeros([BLOCK_M, BLOCK_DMODEL], dtype=tl.float32)
    else:
        # don't have to check boundary for q len
        m_i = tl.load(m_ptrs).to(tl.float32)
        l_i = tl.load(l_ptrs).to(tl.float32)
        acc = tl.load(O_block_ptr).to(tl.float32)

    qk_scale = sm_scale
    qk_scale *= 1.4426950408889634   # 1/log(2)
    # load q: it will stay in SRAM throughout
    if IS_EVEN_M:
        q = tl.load(Q_block_ptr)
    else:
        q = tl.load(Q_block_ptr, boundary_check=(0, 1), padding_option="zero")

    acc, l_i, m_i = _attn_fwd_inner(acc, l_i, m_i, q, K_block_ptr, V_block_ptr, #
                                    start_m, qk_scale, NKV_CTX, #
                                    sliding_window_offset, sliding_window_size,
                                    BLOCK_M, BLOCK_DMODEL, BLOCK_N, SLIDING_WINDOW, IS_EVEN_M, IS_EVEN_N,
                                    COMPLEMENT_SLIDING_WINDOW)
    # epilogue
    if (END):
        m_i += tl.math.log2(l_i)
        acc = acc / l_i[:, None]
    else:
        tl.store(l_ptrs, l_i)

    tl.store(m_ptrs, m_i)
    tl.store(O_block_ptr, acc.to(Out.type.element_ty))

def _forward(
    q, k, v, sm_scale,
    o=None, m=None, l=None, end=False,
    sliding_window=None, init=False,
    complement_sliding_window=False
):
    Lq, Lk, Lv = q.shape[-1], k.shape[-1], v.shape[-1]

    assert Lq == Lk and Lk == Lv
    assert Lk in {16, 32, 64, 128}

    q_round_len = math.ceil(q.shape[2] / 64) * 64

    if sliding_window is not None:
        sliding_window_offset, sliding_window_size = sliding_window
    else:
        sliding_window_offset, sliding_window_size = None, None

    grid = lambda META: (
        triton.cdiv(q.shape[2], META["BLOCK_M"]),
        q.shape[0] * q.shape[1],
    )

    global _BLOCK_N
    global _BLOCK_M

    try:
        with torch.cuda.device(q.device):
            _attn_fwd[grid](
                q, k, v, sm_scale, m, o, l, #
                q.stride(0), q.stride(1), q.stride(2), q.stride(3),  #
                k.stride(0), k.stride(1), k.stride(2), k.stride(3),  #
                v.stride(0), v.stride(1), v.stride(2), v.stride(3),  #
                o.stride(0), o.stride(1), o.stride(2), o.stride(3),  #
                q.shape[0], q.shape[1], k.shape[1], #
                q.shape[2],  #
                q_round_len,
                k.shape[2],
                sliding_window_offset,
                sliding_window_size,
                BLOCK_DMODEL=Lk,  #
                END=end,
                INIT=init,
                BLOCK_M=_BLOCK_M,
                BLOCK_N=_BLOCK_N,
                SLIDING_WINDOW=(sliding_window is not None),
                COMPLEMENT_SLIDING_WINDOW=complement_sliding_window,
                num_warps=4,
                num_stages=4
            )
    except triton.OutOfResources as E:
        _BLOCK_N = _BLOCK_N // 2
        _BLOCK_M = _BLOCK_M // 2
        from warnings import warn
        warn(f"Triton Attention Output Resources. {E}\nUse smaller block size {_BLOCK_N}.")
        with torch.cuda.device(q.device):
            _attn_fwd[grid](
                q, k, v, sm_scale, m, o, l, #
                q.stride(0), q.stride(1), q.stride(2), q.stride(3),  #
                k.stride(0), k.stride(1), k.stride(2), k.stride(3),  #
                v.stride(0), v.stride(1), v.stride(2), v.stride(3),  #
                o.stride(0), o.stride(1), o.stride(2), o.stride(3),  #
                q.shape[0], q.shape[1], k.shape[1], #
                q.shape[2],  #
                q_round_len,
                k.shape[2],
                sliding_window_offset,
                sliding_window_size,
                BLOCK_DMODEL=Lk,  #
                END=end,
                INIT=init,
                BLOCK_M=_BLOCK_M,
                BLOCK_N=_BLOCK_N,
                SLIDING_WINDOW=(sliding_window is not None),
                COMPLEMENT_SLIDING_WINDOW=complement_sliding_window,
                num_warps=4,
                num_stages=4
            )

    if end:
        o = o[:, :, :q.shape[2], :].contiguous().to(q.dtype)

    return o, m, l




##################################################################################################################################################


import torch
import math

# Define the test function for _forward
def test_forward():
    # Define input dimensions
    batch_size = 2
    num_heads = 4
    seq_len = 128
    d_model = 64

    # Create random input tensors
    q = torch.randn((batch_size, num_heads, seq_len, d_model), device='cuda', dtype=torch.float16)
    k = torch.randn((batch_size, num_heads, seq_len, d_model), device='cuda', dtype=torch.float16)
    v = torch.randn((batch_size, num_heads, seq_len, d_model), device='cuda', dtype=torch.float16)

    # Initialize output tensors
    o = torch.zeros_like(q)
    m = torch.zeros((batch_size, num_heads, seq_len), device='cuda', dtype=torch.float32)
    l = torch.zeros((batch_size, num_heads, seq_len), device='cuda', dtype=torch.float32)

    # Define scale and sliding window parameters
    sm_scale = 1.0 / math.sqrt(d_model)

    # Test case 1: sliding_window with complement_sliding_window=False
    sliding_window = (0, 64)
    complement_sliding_window = False
    o1, m1, l1 = _forward(q, k, v, sm_scale, o, m, l, end=True, sliding_window=sliding_window, init=True, complement_sliding_window=complement_sliding_window)

    # Test case 2: sliding_window with complement_sliding_window=True
    complement_sliding_window = True
    o2, m2, l2 = _forward(q, k, v, sm_scale, o, m, l, end=True, sliding_window=sliding_window, init=True, complement_sliding_window=complement_sliding_window)

    # Test case 3: no sliding_window
    sliding_window = None
    o3, m3, l3 = _forward(q, k, v, sm_scale, o, m, l, end=True, sliding_window=sliding_window, init=True, complement_sliding_window=False)

    # Test case 4: init=False
    sliding_window = (0, 64)
    complement_sliding_window = False
    o4, m4, l4 = _forward(q, k, v, sm_scale, o, m, l, end=True, sliding_window=sliding_window, init=False, complement_sliding_window=complement_sliding_window)

    return {
        "test_case_1": (o1, m1, l1),
        "test_case_2": (o2, m2, l2),
        "test_case_3": (o3, m3, l3),
        "test_case_4": (o4, m4, l4)
    }

# Run the tests
result_gold = test_forward()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def attention_fwd_kernel(
    q,
    k,
    v,
    h,
    o,
    s_qh,
    s_qt,
    s_qd,
    s_hh,
    s_ht,
    T,
    scale,
    BT: tl.constexpr,
    BD: tl.constexpr,
    NT: tl.constexpr,
    STORE: tl.constexpr,
    IFCOND: tl.constexpr
):
    i_bh = tl.program_id(0)

    # [BD, BD]
    b_h = tl.zeros([BD, BD], dtype=tl.float32)
    for i in range(0, tl.cdiv(T, BT)):
        p_q = tl.make_block_ptr(q + i_bh * s_qh, (T, BD), (s_qt, s_qd), (i * BT, 0), (BT, BD), (1, 0))
        p_k = tl.make_block_ptr(k + i_bh * s_qh, (BD, T), (s_qd, s_qt), (0, i * BT), (BD, BT), (0, 1))
        p_v = tl.make_block_ptr(v + i_bh * s_qh, (T, BD), (s_qt, s_qd), (i * BT, 0), (BT, BD), (1, 0))
        p_h = tl.make_block_ptr(h + i_bh * s_hh, (NT * BD, BD), (s_ht, s_qd), (i * BD, 0), (BD, BD), (1, 0))
        p_o = tl.make_block_ptr(o + i_bh * s_qh, (T, BD), (s_qt, s_qd), (i * BT, 0), (BT, BD), (1, 0))

        if STORE:
            tl.store(p_h, b_h.to(p_h.dtype.element_ty))
        # [BT, BD]
        b_q = tl.load(p_q)
        b_q = (b_q * scale).to(b_q.dtype)
        # [BD, BT]
        b_k = tl.load(p_k)
        # [BT, BD]
        b_v = tl.load(p_v)

        # [BT, BT]
        b_s = tl.dot(b_q, b_k, allow_tf32=False)
        # [BT, BD]
        b_o = tl.dot(b_s.to(b_q.dtype), b_v, allow_tf32=False)
        if IFCOND:
            if i == 0:
                b_h = tl.dot(b_k, b_v, allow_tf32=False)
            else:
                b_o += tl.dot(b_q, b_h.to(b_q.dtype), allow_tf32=False)
                b_h += tl.dot(b_k, b_v, allow_tf32=False)
        else:
            b_o += tl.dot(b_q, b_h.to(b_q.dtype), allow_tf32=False)
            b_h += tl.dot(b_k, b_v, allow_tf32=False)

        tl.store(p_o, b_o.to(p_o.dtype.element_ty))


class AttentionFunction(torch.autograd.Function):

    @staticmethod
    def forward(ctx, q, k, v, store=False, ifcond=False):
        batch_size, n_heads, seq_len, d_head = q.shape
        scale = d_head ** -0.5
        BD = q.shape[-1] // 2
        BT = 32
        NT = triton.cdiv(seq_len, BT)
        num_stages = 1 #3 if d_head <= 64 else 2
        num_warps = 4

        h = q.new_empty(batch_size, n_heads, NT * BD, BD)
        o = torch.empty_like(q)
        grid = (batch_size * n_heads,)
        attention_fwd_kernel[grid](
            q, k, v, h, o,
            q.stride(1), q.stride(2), q.stride(3), h.stride(1), h.stride(2),
            seq_len, scale,
            BT=BT, BD=BD, NT=NT, STORE=store, IFCOND=ifcond,
            num_warps=num_warps,
            num_stages=num_stages
        )
        return o



##################################################################################################################################################


import torch

def test_attention_fwd():
    # Test 1: Basic Forward Pass (no special conditions)
    B, H, T, D = 2, 8, 1024, 128
    q = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')
    k = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')
    v = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')

    result = AttentionFunction.apply(q, k, v)

    # Test 2: Forward Pass with STORE enabled
    result_store = AttentionFunction.apply(q, k, v, True)

    # Test 3: Forward Pass with IFCOND enabled
    result_ifcond = AttentionFunction.apply(q, k, v, False, True)

    # Test 4: Forward Pass with STORE and IFCOND enabled
    result_store_ifcond = AttentionFunction.apply(q, k, v, True, True)

    # Test 5: Forward Pass with large head size (d_head > 64)
    B, H, T, D = 2, 8, 1024, 128
    q_large_head = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')
    k_large_head = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')
    v_large_head = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')

    result_large_head = AttentionFunction.apply(q_large_head, k_large_head, v_large_head)

    # Test 6: Edge case with smallest sequence length (T = 1)
    B, H, T, D = 2, 8, 1, 128
    q_small_seq = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')
    k_small_seq = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')
    v_small_seq = torch.randn((B, H, T, D), dtype=torch.float, device='cuda')

    result_small_seq = AttentionFunction.apply(q_small_seq, k_small_seq, v_small_seq)

    return {
        "test_case_1": result,
        "test_case_2": result_store,
        "test_case_3": result_ifcond,
        "test_case_4": result_store_ifcond,
        "test_case_5": result_large_head,
        "test_case_6": result_small_seq
    }

# Run all tests
result_gold = test_attention_fwd()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl


@triton.jit
def index_select_cat_bwd_kernel(
    grad_source_ptr,  # *Pointer* to grad_source tensor.
    index_ptr,  # *Pointer* to index tensor.
    grad_output_ptr,  # *Pointer* to grad_output tensor.
    num_rows,
    num_indices,
    num_cols,
    stride0,  # Stride information of input and source tensor.
    stride1,
    BLOCK_SIZE_INDEX: tl.constexpr,  # Number of indices each program should process.
    BLOCK_SIZE_COL: tl.constexpr,  # Number of cols each program should process.
):
    pid0 = tl.program_id(axis=0)  # We use 3D launch grid
    pid1 = tl.program_id(axis=1)

    cols = pid1 * BLOCK_SIZE_COL + tl.arange(0, BLOCK_SIZE_COL)

    # load grad_output
    grad_output_indices = pid0 * BLOCK_SIZE_INDEX + tl.arange(0, BLOCK_SIZE_INDEX)
    grad_output_offsets = (
        grad_output_ptr
        + grad_output_indices[:, None] * stride0
        + cols[None, :] * stride1
    )
    grad_output_mask = (grad_output_indices[:, None] < num_indices) & (
        cols[None, :] < num_cols
    )
    grad_output = tl.load(grad_output_offsets, mask=grad_output_mask).to(tl.float32)

    # select indices from grad_source
    grad_source_indices = tl.load(
        index_ptr + grad_output_indices, mask=(grad_output_indices < num_indices)
    )
    grad_source_offsets = (
        grad_source_ptr
        + grad_source_indices[:, None] * stride0
        + cols[None, :] * stride1
    )

    # compute scaled index add and save
    tl.store(grad_source_offsets, grad_output, mask=grad_output_mask)


def index_select_cat_bwd(
    grad_source: torch.Tensor,
    index: torch.Tensor,
    grad_output: torch.Tensor,
):
    if not (grad_source.is_cuda and grad_output.is_cuda):
        raise ValueError("The grad_source and grad_output tensor must be of type CUDA!")

    if not (grad_source.ndim == 2 and grad_output.ndim == 2):
        raise ValueError(
            f"The grad_source and grad_output must be three-dimensional "
            f"(got {grad_source.ndim} and {grad_output.ndim})!"
        )
    if not grad_source.shape[1] == grad_output.shape[1]:
        raise ValueError(
            f"The number of elements along dimension 1 of grad_source and grad_output must be the same "
            f"(got {grad_source.shape[1]} and {grad_output.shape[1]})"
        )

    num_rows, num_cols = grad_source.shape
    num_indices, num_cols = grad_output.shape
    if not num_rows >= num_indices:
        raise ValueError(
            f"The number of elements along dimension 0 of grad_source must be larger than that of grad_output "
            f"(got {num_rows} and {num_indices})!"
        )
    if not index.shape[0] == num_indices:
        raise ValueError(
            f"The number of indices and the number of elements along dimension 0 of grad_output must match "
            f"(got {index.shape[0]} and {num_indices})!"
        )

    stride0, stride1 = grad_source.stride(0), grad_source.stride(1)
    if not (grad_output.stride(0) == stride0 and grad_output.stride(1) == stride1):
        raise ValueError(
            f"The strides of the grad_source and grad_output tensors must match "
            f"(got {stride0} vs. {grad_output.stride(0)}, {stride1} vs. {grad_output.stride(1)})!"
        )

    def grid(meta):
        return (
            triton.cdiv(num_indices, meta["BLOCK_SIZE_INDEX"]),
            triton.cdiv(num_cols, meta["BLOCK_SIZE_COL"]),
        )

    index_select_cat_bwd_kernel[grid](
        grad_source,
        index,
        grad_output,
        num_rows,
        num_indices,
        num_cols,
        grad_source.stride(0),
        grad_source.stride(1),
        BLOCK_SIZE_INDEX=1,
        BLOCK_SIZE_COL=512,
    )

    return




##################################################################################################################################################


import torch

# Test for index_select_cat_bwd
def test_index_select_cat_bwd():
    results = {}

    # Test case 1: Basic test
    grad_source = torch.zeros(10, 512, device='cuda')
    index = torch.tensor([0, 2, 4, 6, 8], device='cuda')
    grad_output = torch.randn(len(index), grad_source.size(1), device='cuda')
    index_select_cat_bwd(grad_source, index, grad_output)
    results['test_case_1'] = grad_source.clone()

    # Test case 2: Different indices
    grad_source = torch.zeros(10, 512, device='cuda')
    index = torch.tensor([1, 3, 5, 7, 9], device='cuda')
    grad_output = torch.randn(len(index), grad_source.size(1), device='cuda')
    index_select_cat_bwd(grad_source, index, grad_output)
    results['test_case_2'] = grad_source.clone()

    # Test case 3: All indices the same
    grad_source = torch.zeros(10, 512, device='cuda')
    index = torch.tensor([0, 0, 0, 0, 0], device='cuda')
    grad_output = torch.randn(len(index), grad_source.size(1), device='cuda')
    index_select_cat_bwd(grad_source, index, grad_output)
    results['test_case_3'] = grad_source.clone()

    # Test case 4: Maximum index
    grad_source = torch.zeros(10, 512, device='cuda')
    index = torch.tensor([9, 9, 9, 9, 9], device='cuda')
    grad_output = torch.randn(len(index), grad_source.size(1), device='cuda')
    index_select_cat_bwd(grad_source, index, grad_output)
    results['test_case_4'] = grad_source.clone()

    return results

# Run the tests
result_gold = test_index_select_cat_bwd()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.sgmv_expand_slice import _sgmv_expand_slice
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('sgmv_expand_slice', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(4, 12):  # Adjust the range as needed for your testing
            size = 2 ** i
            inputs = torch.rand(size, size, dtype=torch.float32)
            lora_b_weights = torch.rand(1024, size, size, dtype=torch.float16)
            output_tensor = torch.zeros_like(inputs)
            b_seq_start_loc = torch.zeros(size, dtype=torch.int32)
            seq_len_tensor = torch.full((size,), size, dtype=torch.int32)
            lora_indices_tensor = torch.zeros(size, dtype=torch.int32)
            self.input_tensors.append((inputs, lora_b_weights, output_tensor, b_seq_start_loc, seq_len_tensor, lora_indices_tensor))

    def to_cuda(self, input_tensor):
        return tuple(tensor.cuda() for tensor in input_tensor)

    def call_op(self, input_tensor):
        inputs, lora_b_weights, output_tensor, b_seq_start_loc, seq_len_tensor, lora_indices_tensor = input_tensor
        _sgmv_expand_slice(
            inputs,
            lora_b_weights,
            output_tensor,
            b_seq_start_loc,
            seq_len_tensor,
            lora_indices_tensor,
            batches=inputs.size(0),
            max_seq_length=inputs.size(0),
            token_nums=inputs.size(0),
            slice_offset=0,
            slice_size=lora_b_weights.size(-2),
            add_inputs=False
        )
        return output_tensor

    def get_gbps(self, input_tensor, runtime):
        inputs, lora_b_weights, output_tensor, _, _, _ = input_tensor
        total_bytes = inputs.numel() * inputs.element_size() + lora_b_weights.numel() * lora_b_weights.element_size() + output_tensor.numel() * output_tensor.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        inputs, lora_b_weights, _, _, _, _ = input_tensor
        FLOPS = 2 * inputs.size(0) * inputs.size(1) * lora_b_weights.size(-1)
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# Copyright (c) 2023 NVIDIA Corporation & Affiliates. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
Fused Attention
===============
This is a Triton implementation of the Flash Attention algorithm
(see: Dao et al., https://arxiv.org/pdf/2205.14135v2.pdf; Rabe and Staats https://arxiv.org/pdf/2112.05682v2.pdf)
"""
######################################## Imports ######################################## 

# import numpy as np
import pytest
import torch

import triton
import triton.language as tl
######################################## Imports ######################################## 


@triton.jit
def flash_fwd_kernel(Q, K, V, sm_scale,  #
                L, M,  #
                Out,  #
                stride_qz, stride_qh, stride_qm, stride_qk,  #
                stride_kz, stride_kh, stride_kn, stride_kk,  #
                stride_vz, stride_vh, stride_vk, stride_vn,  #
                stride_oz, stride_oh, stride_om, stride_on,  #
                Z, H, N_CTX, D0,  #
                BLOCK_M: tl.constexpr, BLOCK_DMODEL: tl.constexpr, BLOCK_N: tl.constexpr):
    start_m = tl.program_id(0)
    off_hz = tl.program_id(1)

    # TODO: may replace with TMA store without range offset
    # initialize offsets for store
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_n = tl.arange(0, BLOCK_N)
    # initialize pointer to m and l
    m_prev = tl.zeros([BLOCK_M], dtype=tl.float32) - float("inf")
    l_prev = tl.zeros([BLOCK_M], dtype=tl.float32)
    acc = tl.zeros([BLOCK_M, BLOCK_DMODEL], dtype=tl.float32)

    stride_qh_2d = stride_qh // stride_qm // stride_qk

    q_tile_ptr = tl.make_block_ptr(
        base=Q,
        shape=(D0, BLOCK_DMODEL),
        strides=(stride_qm, stride_qk),
        offsets=(off_hz * stride_qh_2d + start_m * BLOCK_M, 0),
        block_shape=(BLOCK_M, BLOCK_DMODEL),
        order=(1, 0),
    )
    k_tile_ptr = tl.make_block_ptr(
        base=K,
        shape=(D0, BLOCK_DMODEL),
        strides=(stride_kn, stride_kk),
        offsets=(off_hz * stride_qh_2d, 0),
        block_shape=(BLOCK_N, BLOCK_DMODEL),
        order=(1, 0),
    )
    v_tile_ptr = tl.make_block_ptr(
        base=V,
        shape=(D0, BLOCK_DMODEL),
        strides=(stride_vk, stride_vn),
        offsets=(off_hz * stride_qh_2d, 0),
        block_shape=(BLOCK_N, BLOCK_DMODEL),
        order=(1, 0),
    )
    out_tile_ptr = tl.make_block_ptr(
        base=Out,
        shape=(D0, BLOCK_DMODEL),
        strides=(stride_om, stride_on),
        offsets=(off_hz * stride_qh_2d + start_m * BLOCK_M, 0),
        block_shape=(BLOCK_M, BLOCK_DMODEL),
        order=(1, 0),
    )
    # load q: it will stay in SRAM throughout
    q = tl.load(q_tile_ptr)

    # loop over k, v and update accumulators
    for start_n in range(0, (start_m + 1) * BLOCK_M, BLOCK_N):
        # -- compute qk ----
        k = tl.load(k_tile_ptr, boundary_check=(0, 1))
        qk = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
        qk += tl.dot(q, tl.trans(k))
        qk *= sm_scale
        qk = tl.where(offs_m[:, None] >= (start_n + offs_n[None, :]), qk, float("-inf"))
        # compute new m
        m_curr = tl.maximum(tl.max(qk, 1), m_prev)
        # correct old l
        l_prev *= tl.exp(m_prev - m_curr)
        # attention weights
        p = tl.exp(qk - m_curr[:, None])
        l_curr = tl.sum(p, 1) + l_prev
        # rescale operands of matmuls
        l_rcp = 1. / l_curr
        p *= l_rcp[:, None]
        acc *= (l_prev * l_rcp)[:, None]
        # update acc
        p = p.to(tl.float16)
        v = tl.load(v_tile_ptr, boundary_check=(0, 1))
        acc += tl.dot(p, v)
        # update m_i and l_i
        l_prev = l_curr
        m_prev = m_curr
        # update pointers
        k_tile_ptr = tl.advance(k_tile_ptr, [BLOCK_N, 0])
        v_tile_ptr = tl.advance(v_tile_ptr, [BLOCK_N, 0])
    # rematerialize offsets to save registers
    start_m = tl.program_id(0)
    offs_m = start_m * BLOCK_M + tl.arange(0, BLOCK_M)
    # write back l and m
    l_ptrs = L + off_hz * N_CTX + offs_m
    m_ptrs = M + off_hz * N_CTX + offs_m
    tl.store(l_ptrs, l_prev)
    tl.store(m_ptrs, m_prev)

    acc = acc.to(tl.float16)
    tl.store(out_tile_ptr, acc, boundary_check=(0, 1))

##################################################################################################################################################  

import numpy as np
import random
import torch 
import os
import pytest
from numpy.random import RandomState
from geak_eval.perf.ROCm.performance_utils_pytest import PytestBenchmarker, do_bench_config, save_all_benchmark_results
from typing import Dict

result_gold = {}

######################################## HELPERS for Eval ######################################## 
def set_seed(seed: int = 42) -> None:
    """
    Set the random seed for reproducibility across multiple libraries and configure PyTorch for deterministic behavior.

    Args:
        seed (int): The seed value to set. Default is 42.
    """
    # Set seed for Python's built-in random module
    random.seed(seed)
    # Set seed for NumPy
    np.random.seed(seed)
    # Set seed for PyTorch on CPU
    torch.manual_seed(seed)
    # Set seed for PyTorch on all GPUs (if available)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    # Ensure deterministic behavior in PyTorch
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set environment variable for hash-based operations
    os.environ['PYTHONHASHSEED'] = str(seed)

######################################## HELPERS for Eval ######################################## 



empty = torch.empty(128, device="cuda")


class _attention(torch.autograd.Function):

    @staticmethod
    def forward(ctx, q, k, v, sm_scale):
        BLOCK = 128
        # shape constraints
        Lq, Lk, Lv = q.shape[-1], k.shape[-1], v.shape[-1]
        assert Lq == Lk and Lk == Lv
        assert Lk in {16, 32, 64, 128}
        o = torch.empty_like(q)
        grid = (triton.cdiv(q.shape[2], BLOCK), q.shape[0] * q.shape[1], 1)
        L = torch.empty((q.shape[0] * q.shape[1], q.shape[2]), device=q.device, dtype=torch.float32)
        m = torch.empty((q.shape[0] * q.shape[1], q.shape[2]), device=q.device, dtype=torch.float32)
        num_warps = 4 if Lk <= 64 else 8
        D0 = q.shape[0] * q.shape[1] * q.shape[2]
        flash_fwd_kernel[grid](
            q, k, v, sm_scale,  #
            L, m,  #
            o,  #
            q.stride(0), q.stride(1), q.stride(2), q.stride(3),  #
            k.stride(0), k.stride(1), k.stride(2), k.stride(3),  #
            v.stride(0), v.stride(1), v.stride(2), v.stride(3),  #
            o.stride(0), o.stride(1), o.stride(2), o.stride(3),  #
            q.shape[0], q.shape[1], q.shape[2], D0,  #
            BLOCK_M=BLOCK, BLOCK_N=BLOCK, BLOCK_DMODEL=Lk,  #
            num_warps=num_warps, num_stages=2)

        ctx.save_for_backward(q, k, v, o, L, m)
        ctx.grid = grid
        ctx.sm_scale = sm_scale
        ctx.BLOCK_DMODEL = Lk
        return o


attention = _attention.apply


@pytest.mark.parametrize('Z, H, N_CTX, D_HEAD', [
    (4, 48, 128, 64),
    (4, 48, 256, 64),
    (4, 48, 512, 64),
    (4, 48, 1024, 64),
    (4, 48, 2048, 64),
    (4, 48, 4096, 64),
    #  (4, 48, 8192, 64), out of memory
])
@pytest.mark.skipif(torch.cuda.get_device_capability()[0] < 9, reason="requires arch 9+")
def test_op(Z, H, N_CTX, D_HEAD, request, dtype=torch.float16):
    
    set_seed()

    q = torch.empty((Z, H, N_CTX, D_HEAD), dtype=dtype, device="cuda").normal_(mean=0.1, std=0.2).requires_grad_()
    k = torch.empty((Z, H, N_CTX, D_HEAD), dtype=dtype, device="cuda").normal_(mean=0.4, std=0.2).requires_grad_()
    v = torch.empty((Z, H, N_CTX, D_HEAD), dtype=dtype, device="cuda").normal_(mean=0.3, std=0.2).requires_grad_()
    sm_scale = 0.2
    dout = torch.randn_like(q)
    # reference implementation
    M = torch.tril(torch.ones((N_CTX, N_CTX), device="cuda"))
    p = torch.matmul(q, k.transpose(2, 3)) * sm_scale
    for z in range(Z):
        for h in range(H):
            p[:, :, M == 0] = float("-inf")
    p = torch.softmax(p.float(), dim=-1).half()
    # p = torch.exp(p)
    ref_out = torch.matmul(p, v)

    # triton implementation
    tri_out = attention(q, k, v, sm_scale)


    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_")
    result_gold[sanitized_key_name] = tri_out.clone().detach().cpu()
    ################################################################### 

    # compare
    torch.testing.assert_close(ref_out, tri_out, atol=1e-2, rtol=0)

def calculate_flash_attention_fwd_tflops(params: dict, ms: float) -> float:
    Z, H, N_CTX, D_HEAD = params['Z'], params['H'], params['N_CTX'], params['D_HEAD']
    flops = 2 * Z * H * N_CTX * N_CTX * D_HEAD 
    tflops = flops / (ms / 1000) / 1e12
    return tflops

def calculate_flash_attention_fwd_gbps(params: dict, ms: float) -> float:
    Z, H, N_CTX, D_HEAD = params['Z'], params['H'], params['N_CTX'], params['D_HEAD']
    dtype_str = params.get('dtype_str', 'fp16') 
    current_dtype = torch.float16
    if dtype_str == 'fp32': current_dtype = torch.float32
    # bf16 was removed from perf test, but keep for completeness if added back
    elif dtype_str == 'bf16': current_dtype = torch.bfloat16 
    element_size = torch.tensor([], dtype=current_dtype).element_size()
    bytes_q, bytes_k, bytes_v, bytes_o_write = [Z * H * N_CTX * D_HEAD * element_size] * 4
    bytes_L_write, bytes_M_write = [Z * H * N_CTX * 4] * 2
    total_bytes = bytes_q + bytes_k + bytes_v + bytes_o_write + bytes_L_write + bytes_M_write
    gbps = total_bytes / (ms / 1000) / 1e9
    return gbps

OP_NAME_FOR_BENCHMARK = "flash_attention_fwd_triton_perf"

# --- NEW Performance Test Function using original test_op's parametrization ---
# It mirrors test_op's parametrize for Z, H, N_CTX, D_HEAD.
# It adds its own parametrize for dtype_str, EXCLUDING bf16 for now.
@pytest.mark.parametrize('Z, H, N_CTX, D_HEAD', [
    (4, 48, 128, 64), (4, 48, 256, 64), (4, 48, 512, 64),
    (4, 48, 1024, 64), (4, 48, 2048, 64), (4, 48, 4096, 64),
    # (2, 12, 1024, 64), (1, 8, 2048, 32) # Example additional shapes if desired
])
@pytest.mark.parametrize('dtype_str', ['fp16']) # MODIFIED: Only fp16 for now to avoid bf16 JIT errors
# @pytest.mark.parametrize('dtype_str', ['fp16', 'fp32']) # Can add fp32 if it works
def test_performance(Z, H, N_CTX, D_HEAD, dtype_str, request): 
    
    cap = torch.cuda.get_device_capability()
    if cap[0] < 9: 
         pytest.skip("Original test requires arch 90+ for this flash attention kernel version.")
    # No bf16 specific skip needed as bf16 is removed from parametrize for this function

    # Shared memory OOM skip for D_HEAD=128 (if such cases were added)
    # This logic relies on knowing the hardcoded BLOCK and num_stages in _attention.forward
    hardcoded_block_in_wrapper = 128 
    hardcoded_num_stages_in_wrapper = 2
    if D_HEAD == 128 and \
       hardcoded_block_in_wrapper == 128 and \
       hardcoded_num_stages_in_wrapper == 2:
        pytest.skip(f"Skipping D_HEAD={D_HEAD} with BLOCK=128, num_stages=2 due to high shared memory demand.")
    
    set_seed()
    
    if dtype_str == 'fp32': current_dtype = torch.float32
    # elif dtype_str == 'bf16': current_dtype = torch.bfloat16 # bf16 removed
    else: current_dtype = torch.float16 

    q = torch.empty((Z, H, N_CTX, D_HEAD), dtype=current_dtype, device="cuda").normal_(mean=0.1, std=0.2).requires_grad_(False)
    k = torch.empty((Z, H, N_CTX, D_HEAD), dtype=current_dtype, device="cuda").normal_(mean=0.4, std=0.2).requires_grad_(False)
    v = torch.empty((Z, H, N_CTX, D_HEAD), dtype=current_dtype, device="cuda").normal_(mean=0.3, std=0.2).requires_grad_(False)
    sm_scale = 0.2 

    op_lambda = lambda: attention(q, k, v, sm_scale)

    bench_config = do_bench_config(warm_up=10, repetition=50) 
    benchmarker = PytestBenchmarker(op_callable=op_lambda,
                                    op_name=OP_NAME_FOR_BENCHMARK,
                                    config=bench_config)
    current_params_for_logs_and_calc = {
        "Z": Z, "H": H, "N_CTX": N_CTX, "D_HEAD": D_HEAD,
        "dtype_str": dtype_str, "sm_scale": sm_scale
    }
    
    perf_result = benchmarker.run_benchmark(current_params_dict=current_params_for_logs_and_calc,
                                            gbps_calculator=calculate_flash_attention_fwd_gbps,
                                            tflops_calculator=calculate_flash_attention_fwd_tflops)


######################################## HELPERS for Eval ########################################     
# --- Pytest hook to save the dictionary at the end of the session ---  
def test_save_results():  
    """  
    Called after whole test run finished, right before returning the exit status to the system.  
    """
    print('Inside session finish...')
    if "_CALL_SUCCESS_" not in result_gold:
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[0.0]])
    OUTPUT_FILENAME = __file__.replace('.','_') + '.pt'
    print(f"\nSaving all y_triton results to {OUTPUT_FILENAME}...")  
    # Ensure the directory for the output file exists if it's in a subdirectory  
    output_dir = os.path.dirname(OUTPUT_FILENAME)  
    if output_dir and not os.path.exists(output_dir):  
        os.makedirs(output_dir, exist_ok=True)  
    torch.save(result_gold, OUTPUT_FILENAME)       
    print(f"Successfully saved {len(result_gold)} y_triton tensors to {OUTPUT_FILENAME}.")  

def test_save_performance_results():
    """
    Called after the test_performance function finishes.
    This is a separate hook to ensure performance results are saved.
    """
    print('\nPytest session finishing... Saving benchmark results...')

    output_directory = os.path.join(os.path.dirname(__file__), "perf")  # Save in a "perf" subdirectory next to the test file
    os.makedirs(output_directory, exist_ok=True)
    
    save_all_benchmark_results(output_directory)
    print(f"All benchmark results attempted to save to: {output_directory}")
######################################## HELPERS for Eval ########################################

try:
    from flash_attn.flash_attn_interface import flash_attn_func
    HAS_FLASH = True
except BaseException:
    HAS_FLASH = False

BATCH, N_HEADS, N_CTX, D_HEAD = 4, 48, 4096, 64
# vary seq length for fixed head and batch=4
configs = [
    triton.testing.Benchmark(
        x_names=['N_CTX'],
        x_vals=[2**i for i in range(10, 14)],
        line_arg='provider',
        line_vals=['triton'] + (['flash'] if HAS_FLASH else []),
        line_names=['Triton'] + (['Flash'] if HAS_FLASH else []),
        styles=[('red', '-'), ('blue', '-')],
        ylabel='ms',
        plot_name=f'fused-attention-batch{BATCH}-head{N_HEADS}-d{D_HEAD}-{mode}',
        args={
            'H': N_HEADS,
            'BATCH': BATCH,
            'D_HEAD': D_HEAD,
            'dtype': torch.float16,
            'mode': mode,
        },
    ) for mode in ['fwd', 'bwd']
]


@triton.testing.perf_report(configs)
def bench_flash_attention(BATCH, H, N_CTX, D_HEAD, mode, provider, dtype=torch.float16, device="cuda"):
    assert mode in ['fwd', 'bwd']
    if provider == "triton":
        q = torch.randn((BATCH, H, N_CTX, D_HEAD), dtype=dtype, device="cuda", requires_grad=True)
        k = torch.randn((BATCH, H, N_CTX, D_HEAD), dtype=dtype, device="cuda", requires_grad=True)
        v = torch.randn((BATCH, H, N_CTX, D_HEAD), dtype=dtype, device="cuda", requires_grad=True)
        sm_scale = 1.3
        fn = lambda: attention(q, k, v, sm_scale)
        if mode == 'bwd':
            o = fn()
            do = torch.randn_like(o)
            fn = lambda: o.backward(do, retain_graph=True)
        ms = triton.testing.do_bench(fn)
        return ms
    if provider == "flash":
        lengths = torch.full((BATCH, ), fill_value=N_CTX, device=device)
        cu_seqlens = torch.zeros((BATCH + 1, ), device=device, dtype=torch.int32)
        cu_seqlens[1:] = lengths.cumsum(0)
        qkv = torch.randn((BATCH * N_CTX, 3, H, D_HEAD), dtype=dtype, device=device, requires_grad=True)
        fn = lambda: flash_attn_func(qkv, cu_seqlens, 0., N_CTX, causal=True)
        if mode == 'bwd':
            o = fn()
            do = torch.randn_like(o)
            fn = lambda: o.backward(do, retain_graph=True)
        ms = triton.testing.do_bench(fn)
        return ms





# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def _dropout(
    x_ptr,  # pointer to the input
    x_keep_ptr,  # pointer to a mask of 0s and 1s
    output_ptr,  # pointer to the output
    n_elements,  # number of elements in the `x` tensor
    p,  # probability that an element of `x` is changed to zero
    BLOCK_SIZE: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    # Load data
    x = tl.load(x_ptr + offsets, mask=mask)
    x_keep = tl.load(x_keep_ptr + offsets, mask=mask)
    # The line below is the crucial part, described in the paragraph above!
    output = tl.where(x_keep, x / (1 - p), 0.0)
    # Write-back output
    tl.store(output_ptr + offsets, output, mask=mask)


def dropout(x, x_keep, p):
    output = torch.empty_like(x)
    assert x.is_contiguous()
    n_elements = x.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta['BLOCK_SIZE']), )
    _dropout[grid](x, x_keep, output, n_elements, p, BLOCK_SIZE=1024)
    return output



##################################################################################################################################################


# Test for the dropout function
def test_dropout():
    # Dictionary to store test case results
    results = {}
    
    # Test case 1
    x = torch.randn(size=(10,)).cuda()
    p = 0.5
    x_keep = (torch.rand(size=(10,)) > p).to(torch.int32).cuda()
    output = dropout(x, x_keep=x_keep, p=p)
    results['test_case_1'] = output

    # Test case 2: p = 0 (no dropout)
    p = 0.0
    x_keep = (torch.rand(size=(10,)) > p).to(torch.int32).cuda()
    output = dropout(x, x_keep=x_keep, p=p)
    results['test_case_2'] = output

    # Test case 3: p = 1 (full dropout)
    p = 1.0
    x_keep = (torch.rand(size=(10,)) > p).to(torch.int32).cuda()
    output = dropout(x, x_keep=x_keep, p=p)
    results['test_case_3'] = output

    # Test case 4: different block size
    p = 0.5
    x_keep = (torch.rand(size=(10,)) > p).to(torch.int32).cuda()
    output = dropout(x, x_keep=x_keep, p=p)
    results['test_case_4'] = output
    
    return results

# Run tests and store result
result_gold = test_dropout()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import time
import torch

import triton
import triton.language as tl


@triton.autotune(
    configs=[
        triton.Config({}, num_stages=2, num_warps=8),
        triton.Config({}, num_stages=2, num_warps=4),
        triton.Config({}, num_stages=2, num_warps=2),
        triton.Config({}, num_stages=2, num_warps=1),
     ],
    key=['K'],
)
@triton.jit
def quantize_int8_perrow_kernel(
    fpa_ptr, a_ptr, as_ptr,
    M, K, 
    stride_fpam, stride_fpak,
    stride_am, stride_ak,
    stride_asm,
    # Meta-parameters
    BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
):
    pid_m = tl.program_id(axis=0)
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M

    fpa_ptrs = fpa_ptr + offs_am[:, None] * stride_fpam + offs_k[None, :] * stride_fpak
    a_ptrs = a_ptr + offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak
    a_max = tl.zeros((BLOCK_SIZE_M,), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        fpa = tl.load(fpa_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K, other=0.0)
        a_max = tl.maximum(a_max, tl.max(tl.abs(fpa), axis=1))
        fpa_ptrs += BLOCK_SIZE_K * stride_fpak
    a_scale = (a_max / 127.)
    fpa_ptrs = fpa_ptr + offs_am[:, None] * stride_fpam + offs_k[None, :] * stride_fpak
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        fpa = tl.load(fpa_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K, other=0.0)
        inta = (fpa / a_scale[:, None]).to(tl.int8)
        tl.store(a_ptrs, inta, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K)
        fpa_ptrs += BLOCK_SIZE_K * stride_fpak
        a_ptrs += BLOCK_SIZE_K * stride_ak
    as_offs = pid_m * BLOCK_SIZE_M * stride_asm + tl.arange(0, BLOCK_SIZE_M)
    tl.store(as_ptr + as_offs, a_scale)


def quantize_int8_perrow(fpa):
    a = torch.empty(fpa.shape, device=fpa.device, dtype=torch.int8)
    a_scale = torch.empty(fpa.shape[0], device=fpa.device, dtype=torch.float16)
    M, K = fpa.shape
    BLOCK_SIZE_M = 1
    BLOCK_SIZE_K = triton.next_power_of_2(K)
    grid = (M // BLOCK_SIZE_M,)
    quantize_int8_perrow_kernel[grid](
        fpa, a, a_scale,
        M, K,
        fpa.stride(0), fpa.stride(1),
        a.stride(0), a.stride(1),
        a_scale.stride(0),
        BLOCK_SIZE_M, BLOCK_SIZE_K,
    )
    return a, a_scale


@triton.autotune(
    configs=[
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 8}, num_stages=3, num_warps=8),
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64,  'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 64,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64,  'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 32,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64,  'BLOCK_SIZE_N': 32,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
        triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 32,  'BLOCK_SIZE_N': 64,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32,  'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 64,  'GROUP_SIZE_M': 8}, num_stages=3, num_warps=8),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 8}, num_stages=2, num_warps=4),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32,  'GROUP_SIZE_M': 16}, num_stages=4, num_warps=4),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 64,  'GROUP_SIZE_M': 16}, num_stages=3, num_warps=8),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 16}, num_stages=2, num_warps=4),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 64,  'GROUP_SIZE_M': 16}, num_stages=4, num_warps=4),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 16}, num_stages=3, num_warps=8),
	    triton.Config({'SPLIT_K': 1, 'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 256, 'GROUP_SIZE_M': 16}, num_stages=2, num_warps=4),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 64, 'GROUP_SIZE_M': 8}, num_stages=3, num_warps=8),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64,  'BLOCK_SIZE_N': 256, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 64,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64,  'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 32,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64,  'BLOCK_SIZE_N': 32,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 32,  'BLOCK_SIZE_N': 64,  'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32,  'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
		triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 64,  'GROUP_SIZE_M': 8}, num_stages=3, num_warps=8),
        triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 8}, num_stages=2, num_warps=4),
		triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 32,  'GROUP_SIZE_M': 16}, num_stages=4, num_warps=4),
		triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 64,  'GROUP_SIZE_M': 16}, num_stages=3, num_warps=8),
		triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 16}, num_stages=2, num_warps=4),
		triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 64,  'GROUP_SIZE_M': 16}, num_stages=4, num_warps=4),
		triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 64, 'BLOCK_SIZE_N': 64, 'BLOCK_SIZE_K': 128, 'GROUP_SIZE_M': 16}, num_stages=3, num_warps=8),
		triton.Config({'SPLIT_K': 2, 'BLOCK_SIZE_M': 32, 'BLOCK_SIZE_N': 32, 'BLOCK_SIZE_K': 256, 'GROUP_SIZE_M': 16}, num_stages=2, num_warps=4),
    ],
    key=['M', 'N', 'K'],
    reset_to_zero=['c_ptr']
)
@triton.jit
def matmul_kernel(
    # Pointers to matrices
    a_ptr, as_ptr, b_ptr, bs_ptr, c_ptr,
    # Matrix dimensions
    M, N, K,
    # The stride variables represent how much to increase the ptr by when moving by 1
    # element in a particular dimension. E.g. `stride_am` is how much to increase `a_ptr`
    # by to get the element one row down (A has M rows).
    stride_am, stride_ak,
    stride_asm,
    stride_bk, stride_bn,
    stride_bsn,
    stride_cm, stride_cn,
    # Meta-parameters
    BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr, SPLIT_K: tl.constexpr, 
):
    """Kernel for computing the matmul C = A x B.
    A has shape (M, K), B has shape (K, N) and C has shape (M, N)
    """
    # -----------------------------------------------------------
    # Map program ids `pid` to the block of C it should compute.
    # This is done in a grouped ordering to promote L2 data reuse.
    # See above `L2 Cache Optimizations` section for details.
    pid = tl.program_id(axis=0)
    pid_sp_k = tl.program_id(axis=1)
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (pid % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m

    # ----------------------------------------------------------
    # Create pointers for the first blocks of A and B.
    # We will advance this pointer as we move in the K direction
    # and accumulate
    # `a_ptrs` is a block of [BLOCK_SIZE_M, BLOCK_SIZE_K] pointers
    # `b_ptrs` is a block of [BLOCK_SIZE_K, BLOCK_SIZE_N] pointers
    # See above `Pointer Arithmetics` section for details
    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = pid_sp_k * BLOCK_SIZE_K + tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)
    as_ptrs = as_ptr + offs_am * stride_asm
    bs_ptrs = bs_ptr + offs_bn * stride_bsn
    a_scale = tl.load(as_ptrs, mask=offs_am < M, other=0.0)
    b_scale = tl.load(bs_ptrs, mask=offs_bn < N, other=0.0)
    # -----------------------------------------------------------
    # Iterate to compute a block of the C matrix.
    # We accumulate into a `[BLOCK_SIZE_M, BLOCK_SIZE_N]` block
    # of fp32 values for higher accuracy.
    # `accumulator` will be converted back to fp16 after the loop.
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.int32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K * SPLIT_K)):
        # Load the next block of A and B, generate a mask by checking the K dimension.
        # If it is out of bounds, set it to 0.
        a = tl.load(a_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K * SPLIT_K, other=0.0)
        b = tl.load(b_ptrs, mask=offs_k[:, None] < K - k * BLOCK_SIZE_K * SPLIT_K, other=0.0)
        # We accumulate along the K dimension.
        accumulator += tl.dot(a, b)
        # Advance the ptrs to the next K block.
        a_ptrs += BLOCK_SIZE_K * SPLIT_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * SPLIT_K * stride_bk
    # You can fuse arbitrary activation functions here
    # while the accumulator is still in FP32!
    c = (accumulator.to(tl.float32) * a_scale[:, None] * b_scale[None, :]).to(tl.float16)
    # -----------------------------------------------------------
    # Write back the block of the output matrix C with masks.
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + stride_cm * offs_cm[:, None] + stride_cn * offs_cn[None, :]
    c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)
    if SPLIT_K == 1:
        tl.store(c_ptrs, c, mask=c_mask)
    else:
        tl.atomic_add(c_ptrs, c, mask=c_mask)


def matmul_quantize_int8(fpa, b, b_scale, out=None):
    a, a_scale = quantize_int8_perrow(fpa)
    # a, a_scale = quantize_int8(fpa, axis=1)
    return matmul_int8(a, a_scale, b, b_scale, out)


def matmul_int8(a, a_scale, b, b_scale, out=None):
    # Check constraints.
    assert a.shape[1] == b.shape[0], "Incompatible dimensions"
    M, K = a.shape
    K, N = b.shape
    # Allocates output.
    if out == None:
        c = torch.zeros((M, N), device=a.device, dtype=torch.float16)
    else:
        c = out.fill_(0.)
    grid = lambda META: (
        triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']),
        META['SPLIT_K'],
    )
    matmul_kernel[grid](
        a, a_scale, b, b_scale, c,
        M, N, K,
        a.stride(0), a.stride(1),
        a_scale.stride(0),
        b.stride(0), b.stride(1),
        b_scale.stride(0),
        c.stride(0), c.stride(1),
    )
    return c


def quantize_int8(weight, axis=0):
    # Weight shape: [H1, H2]
    # Scale shape: [H2]
    scale = weight.abs().amax(axis, keepdim=True) / 127.
    weight = (weight / scale).to(torch.int8)
    # col major will accelerate i8xi8 kernel.
    if axis == 0:
        weight = weight.t().contiguous().t()
    scale = scale.squeeze(axis)
    return weight, scale, None




##################################################################################################################################################


# Test case for quantize_int8 and matmul_quantize_int8
def test_quantize_and_matmul():
    # Initialize test results dictionary
    test_results = {}

    # Generate random float matrix A (fpa) and B
    M, K, N = 128, 64, 128  # Example sizes for A, B, and C
    device = 'cuda'  # Ensure this runs on GPU

    # Test case 1
    fpa = torch.randn((M, K), device=device, dtype=torch.float32)
    b = torch.randn((K, N), device=device, dtype=torch.float32)
    b, b_scale, _ = quantize_int8(b)
    c_quantized = matmul_quantize_int8(fpa, b, b_scale)
    test_results['test_case_1'] = c_quantized

    # Test case 2
    M, K, N = 32, 16, 32
    fpa = torch.randn((M, K), device=device, dtype=torch.float32)
    b = torch.randn((K, N), device=device, dtype=torch.float32)
    b, b_scale, _ = quantize_int8(b)
    c_quantized = matmul_quantize_int8(fpa, b, b_scale)
    test_results['test_case_2'] = c_quantized

    return test_results

# Run the test case
result_gold = test_quantize_and_matmul()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def _add_kernel(A, B, C, size, BLOCK: tl.constexpr):
    """add kernel."""
    prog_id = tl.program_id(0)
    offs = prog_id * BLOCK + tl.arange(0, BLOCK)
    a = tl.load(A + offs, mask=offs < size)
    b = tl.load(B + offs, mask=offs < size)
    tl.store(C + offs, a + b, mask=offs < size)

def custom_add(a, b):
    """custom add one."""
    c = torch.empty_like(a)
    size = c.size(0)
    BLOCK = 16

    grid = (triton.cdiv(size, BLOCK), )
    _add_kernel[grid](a, b, c, size, BLOCK=BLOCK)
    return c




##################################################################################################################################################


import torch

def test_add():
    # 测试用例 1：简单的两个向量加法
    a = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16], dtype=torch.float32, device='cuda')
    b = torch.tensor([16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1], dtype=torch.float32, device='cuda')
    c = custom_add(a, b)
    test_results = {
        "test_case_1": c,
    }
    # 测试用例 2：不同值的加法
    a = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8], dtype=torch.float32, device='cuda')
    b = torch.tensor([8, 7, 6, 5, 4, 3, 2, 1], dtype=torch.float32, device='cuda')
    c = custom_add(a, b)
    test_results["test_case_2"] = c
    # 测试用例 3：更大向量的加法
    a = torch.arange(32, dtype=torch.float32, device='cuda')
    b = torch.arange(32, 0, -1, dtype=torch.float32, device='cuda')
    c = custom_add(a, b)
    test_results["test_case_3"] = c
    # 测试用例 4：空向量的边界情况
    a = torch.tensor([], dtype=torch.float32, device='cuda')
    b = torch.tensor([], dtype=torch.float32, device='cuda')
    c = custom_add(a, b)
    test_results["test_case_4"] = c
    # test_results = {
    #     "test_case_1": custom_add(a, b),
    #     "test_case_2": custom_add(a, b),
    #     "test_case_3": custom_add(a, b),
    #     "test_case_4": custom_add(a, b),
    # }
    return test_results

result_gold = test_add()


# Copyright(C) [2025] Advanced Micro Devices, Inc. All rights reserved.
import argparse  
import sys  
import pytest  
  
import torch  
import triton  
import triton.language as tl  
import os  
import json  
import math  
from itertools import product  
  
######################################## HELPERS utils ######################################## 
def is_cuda():  
    return triton.runtime.driver.active.get_current_target().backend == "cuda"  
  
  
def is_hip():  
    return triton.runtime.driver.active.get_current_target().backend == "hip"  
  
  
def get_cuda_autotune_config():  
    return [  
        triton.Config({}, num_warps=4, num_stages=1),  
        triton.Config({}, num_warps=8, num_stages=1),  
        triton.Config({}, num_warps=16, num_stages=1),  
    ]  
  
  
def get_hip_autotune_config():  
    return [  
        triton.Config({'waves_per_eu': we}, num_warps=wa, num_stages=1) for we, wa in product([1, 2, 4], [4, 8, 16])  
    ]  
  
  
def get_autotune_config():  
    if is_cuda():  
        return get_cuda_autotune_config()  
    else:  
        return get_hip_autotune_config()  
  
######################################## HELPERS utils ######################################## 


@triton.autotune(configs=get_autotune_config(), key=['n_rows', 'n_cols'], use_cuda_graph=True)  
@triton.jit  
def layernorm_kernel(x_ptr, y_ptr, w_ptr, b_ptr, mean_ptr, rstd_ptr, x_row_stride, y_row_stride, n_rows, n_cols, eps,  
                     BLOCK_SIZE: tl.constexpr):  
  
    #program id  
    row = tl.program_id(0)  
    x_ptr_start = x_ptr + (row * x_row_stride)  
    y_ptr_start = y_ptr + (row * y_row_stride)  
  
    loop_num = tl.cdiv(n_cols, BLOCK_SIZE) - 1  
  
    #calculate mean  
    mean = 0  
    _mean = tl.zeros([BLOCK_SIZE], dtype=tl.float32)  
    loop_num_l = loop_num  
    for b in range(0, loop_num_l):  
        col_offsets = b * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)  
        x_block = tl.load(x_ptr_start + col_offsets).to(tl.float32)  #Unmasked loads  
        _mean += x_block  
  
    #For last iteration, do masked load  
    col_offsets = loop_num_l * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)  
    x_block = tl.load(x_ptr_start + col_offsets, mask=col_offsets < n_cols, other=0.).to(tl.float32)  
    _mean += x_block  
    mean = tl.sum(_mean, axis=0) / n_cols  
  
    #variance  
    _var = tl.zeros([BLOCK_SIZE], dtype=tl.float32)  
    loop_num_l = loop_num  
    for b in range(0, loop_num_l):  
        col_offsets = b * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)  
        x_block = tl.load(x_ptr_start + col_offsets).to(tl.float32)  #Unmasked loads  
        x_block = x_block - mean  
        _var += x_block * x_block  
  
    #For last iteration, do masked load  
    col_offsets = loop_num_l * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)  
    x_block = tl.load(x_ptr_start + col_offsets, mask=col_offsets < n_cols, other=0.).to(tl.float32)  
    x_block = tl.where(col_offsets < n_cols, x_block - mean, 0.)  
    _var += x_block * x_block  
  
    var = tl.sum(_var, axis=0) / n_cols  
    rstd = tl.rsqrt(var + eps)  
  
    # Write mean / rstd  
    tl.store(mean_ptr + row, mean)  
    tl.store(rstd_ptr + row, rstd)  
  
    #Normalize and store  
    loop_num_l = loop_num  
    for b in range(0, loop_num_l):  
        col_offsets = b * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)  
        w_block = tl.load(w_ptr + col_offsets)  
        b_block = tl.load(b_ptr + col_offsets)  
        x_block = tl.load(x_ptr_start + col_offsets).to(tl.float32)  
        y_block = (x_block - mean) * rstd  
        y_block = y_block * w_block + b_block  
        tl.store(y_ptr_start + col_offsets, y_block)  
  
    #For last iteration, do masked load and store  
    col_offsets = loop_num_l * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)  
    mask = col_offsets < n_cols  
    w_block = tl.load(w_ptr + col_offsets, mask=mask, other=0.0)  
    b_block = tl.load(b_ptr + col_offsets, mask=mask, other=0.0)  
    x_block = tl.load(x_ptr_start + col_offsets, mask=mask, other=0.0).to(tl.float32)  
    y_block = (x_block - mean) * rstd  
    y_block = y_block * w_block + b_block  
    tl.store(y_ptr_start + col_offsets, y_block, mask=mask)  
  
  
##################################################################################################################################################  

import numpy as np
import random
import torch 
import os
from geak_eval.perf.ROCm.performance_utils_pytest import PytestBenchmarker, do_bench_config, save_all_benchmark_results
from typing import Dict


result_gold = {}

######################################## HELPERS for Eval ######################################## 
def set_seed(seed: int = 42) -> None:
    """
    Set the random seed for reproducibility across multiple libraries and configure PyTorch for deterministic behavior.

    Args:
        seed (int): The seed value to set. Default is 42.
    """
    # Set seed for Python's built-in random module
    random.seed(seed)
    # Set seed for NumPy
    np.random.seed(seed)
    # Set seed for PyTorch on CPU
    torch.manual_seed(seed)
    # Set seed for PyTorch on all GPUs (if available)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    # Ensure deterministic behavior in PyTorch
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # Set environment variable for hash-based operations
    os.environ['PYTHONHASHSEED'] = str(seed)

######################################## HELPERS for Eval ######################################## 
  

@triton.jit  
def _layer_norm_bwd_dx_fused(DX,  # pointer to the input gradient  
                             DY,  # pointer to the output gradient  
                             DW,  # pointer to the partial sum of weights gradient  
                             DB,  # pointer to the partial sum of biases gradient  
                             X,  # pointer to the input  
                             W,  # pointer to the weights  
                             Mean,  # pointer to the mean  
                             Rstd,  # pointer to the 1/std  
                             stride,  # how much to increase the pointer when moving by 1 row  
                             N,  # number of columns in X  
                             NUM_ROWS: tl.constexpr, BLOCK_SIZE_N: tl.constexpr):  
    # Map the program id to the elements of X, DX, and DY it should compute.  
    pid = tl.program_id(0)  
    pid_n = tl.program_id(1)  
    tile_num = tl.num_programs(0)  
    rows_per_tile = NUM_ROWS // tile_num  
    if pid < NUM_ROWS % tile_num:  
        rows_per_tile += 1  
  
    cols = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)  
    mask = cols < N  
    row = pid  
    for _ in range(0, rows_per_tile):  
        x_ptrs = X + row * stride  
        dy_ptrs = DY + row * stride  
        dx_ptrs = DX + row * stride  
        dw_ptrs = DW + pid * N + cols  
        db_ptrs = DB + pid * N + cols  
        # Load data to SRAM  
        x = tl.load(x_ptrs + cols, mask=mask, other=0).to(tl.float32)  
        dy = tl.load(dy_ptrs + cols, mask=mask, other=0).to(tl.float32)  
        w = tl.load(W + cols, mask=mask).to(tl.float32)  
        mean = tl.load(Mean + row)  
        rstd = tl.load(Rstd + row)  
        # Compute dx  
        xhat = (x - mean) * rstd  
        wdy = w * dy  
        xhat = tl.where(mask, xhat, 0.)  
        wdy = tl.where(mask, wdy, 0.)  
        c1 = tl.sum(xhat * wdy, axis=0) / N  
        c2 = tl.sum(wdy, axis=0) / N  
        dx = (wdy - (xhat * c1 + c2)) * rstd  
        # Write dx  
        tl.store(dx_ptrs + cols, dx, mask=mask)  
        # Accumulate partial sums for dw/db  
        partial_dw = (dy * xhat).to(w.dtype)  
        partial_db = (dy).to(w.dtype)  
        partial_dw += tl.load(dw_ptrs, mask=mask)  
        partial_db += tl.load(db_ptrs, mask=mask)  
        tl.store(dw_ptrs, partial_dw, mask=mask)  
        tl.store(db_ptrs, partial_db, mask=mask)  
        row += tile_num  
  
  
@triton.jit  
def _layer_norm_bwd_dwdb(DW,  # pointer to the partial sum of weights gradient  
                         DB,  # pointer to the partial sum of biases gradient  
                         FINAL_DW,  # pointer to the weights gradient  
                         FINAL_DB,  # pointer to the biases gradient  
                         M,  # GROUP_SIZE_M  
                         N,  # number of columns  
                         BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr):  
    # Map the program id to the elements of DW and DB it should compute.  
    pid = tl.program_id(0)  
    cols = pid * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)  
    dw = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)  
    db = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)  
    # Iterate through the rows of DW and DB to sum the partial sums.  
    for i in range(0, M, BLOCK_SIZE_M):  
        rows = i + tl.arange(0, BLOCK_SIZE_M)  
        mask = (rows[:, None] < M) & (cols[None, :] < N)  
        offs = rows[:, None] * N + cols[None, :]  
        dw += tl.load(DW + offs, mask=mask, other=0.)  
        db += tl.load(DB + offs, mask=mask, other=0.)  
    # Write the final sum to the output.  
    sum_dw = tl.sum(dw, axis=0)  
    sum_db = tl.sum(db, axis=0)  
    tl.store(FINAL_DW + cols, sum_dw, mask=cols < N)  
    tl.store(FINAL_DB + cols, sum_db, mask=cols < N)  



class LayerNorm(torch.autograd.Function):  
  
    @staticmethod  
    def forward(ctx, x, normalized_shape, weight, bias, eps=1e-5):  
        y = torch.empty_like(x)  
        M, N = x.shape  
        mean = torch.empty((M, ), dtype=torch.float32, device=x.device)  
        rstd = torch.empty((M, ), dtype=torch.float32, device=x.device)  
        # Less than 64KB per feature: enqueue fused kernel  
        MAX_FUSED_SIZE = 65536 // x.element_size()  
        BLOCK_SIZE = min(MAX_FUSED_SIZE, triton.next_power_of_2(N))  
        # heuristics for number of warps  
        num_warps = min(max(BLOCK_SIZE // 256, 1), 8)  
        layernorm_kernel[(M, )](x, y, weight, bias, mean, rstd, x.stride(0), y.stride(0), M, N, eps, BLOCK_SIZE)  
        ctx.save_for_backward(x, weight, bias, mean, rstd)  
        ctx.BLOCK_SIZE = BLOCK_SIZE  
        ctx.num_warps = num_warps  
        ctx.eps = eps  
  
        return y  
  
    @staticmethod  
    def backward(ctx, dy):  
        x, w, b, m, v = ctx.saved_tensors  
        N = w.shape[0]  
        x_arg = x.reshape(-1, x.shape[-1])  
        M = x_arg.shape[0]  
        tile_num = max(min(256, M // 4), 1)  
        # allocate output  
        _dw = torch.zeros((tile_num, N), dtype=x.dtype, device=w.device)  
        _db = torch.zeros((tile_num, N), dtype=x.dtype, device=w.device)  
        dw = torch.empty((N, ), dtype=w.dtype, device=w.device)  
        db = torch.empty((N, ), dtype=w.dtype, device=w.device)  
        dx = torch.empty_like(dy)  
  
        # enqueue kernel using forward pass heuristics  
        # also compute partial sums for DW and DB  
        M, N = x_arg.shape  
        grid_bwd = lambda meta: (tile_num, triton.cdiv(N, meta['BLOCK_SIZE_N']))  
        _layer_norm_bwd_dx_fused[grid_bwd](  #  
            dx, dy, _dw, _db, x, w, m, v,  #  
            x_arg.stride(0), N,  #  
            NUM_ROWS=M,  #  
            BLOCK_SIZE_N=ctx.BLOCK_SIZE,  #  
            num_warps=ctx.num_warps)  
        grid_reduce = lambda meta: [triton.cdiv(N, meta['BLOCK_SIZE_N'])]  
        # accumulate partial sums in separate kernel  
        _layer_norm_bwd_dwdb[grid_reduce](  
            _dw, _db, dw, db, min(tile_num, M), N,  #  
            BLOCK_SIZE_M=32,  #  
            BLOCK_SIZE_N=128)  
  
        return dx, None, dw, db, None  
  
  
layernorm = LayerNorm.apply  
  
  
def torch_layernorm(x, w_shape, w, b):  
    M, N = x.shape  
    y_torch = torch.nn.functional.layer_norm(x, w_shape, w, b, eps=1e-5)  
    return y_torch  
  
  
def run_layernorm(M, N):  
    print(f"Running Layernorm on shape ({M},{N})")  
    set_seed()  
    x = torch.randn(M, N, device='cuda')  
    w_shape = (N, )  
    w = torch.rand(w_shape, device='cuda')  
    b = torch.rand(w_shape, device='cuda')  
    y_triton = layernorm(x, w_shape, w, b)  
  
    return y_triton  
  
  
#pytest  
@pytest.mark.parametrize('M, N', [(1823, 781), (2, 128), (1, 4), (128, 2), (1, 128), (8192, 8192), (4096, 8192),  
                                  (359, 1), (1, 359), (1, 131072), (1, 89999)])  
def test_layernorm(M, N, request, eps=1e-5):  
    set_seed()  
    x = torch.randn(M, N, device='cuda')  
    w_shape = (N, )  
    w = torch.rand(w_shape, device='cuda', requires_grad=True)  
    b = torch.rand(w_shape, device='cuda', requires_grad=True)  
  
    dy = 0.1 * torch.randn_like(x)  
    x.requires_grad_(True)  
  
    # forward pass  
    y_triton = layernorm(x, w_shape, w, b, eps)  
    y_ref = torch.nn.functional.layer_norm(x, w_shape, w, b, eps)  

    result_gold['_CALL_SUCCESS_'] = torch.tensor([[1.0]])
    ################### save tri_out in result_gold ###################
    test_case_name = request.node.name
    sanitized_key_name = test_case_name.replace("::", "_").replace("[", "_").replace("]", "").replace("-", "_") + "_fwd"
    result_gold[sanitized_key_name] = y_triton.clone().detach().cpu()
    ###################################################################
  
    # backward pass (triton)  
    y_triton.backward(dy, retain_graph=True)  
    dx_tri, dw_tri, db_tri = [_.grad.clone() for _ in [x, w, b]]  
    x.grad, w.grad, b.grad = None, None, None  
  
    #backward pass (torch)  
    y_ref.backward(dy, retain_graph=True)


# --- Define TFLOPS and GB/s calculators for LayerNorm Forward ---
def calculate_layernorm_gbps(params: dict, ms: float) -> float:
    M = params['M']
    N = params['N']
    # Assuming params contains dtype_str for x, w, b or we infer.
    # For layernorm:
    # Read x (M*N), w (N), b (N)
    # Write y (M*N), mean (M), rstd (M)
    try:
        # Assuming x, y, w, b, mean, rstd are of similar precision for byte counting
        # A more precise calculation would use actual dtypes if they vary significantly.
        dtype = torch.float16 # Default assumption if not in params
        if 'dtype_str' in params:
            if params['dtype_str'] == 'fp32': dtype = torch.float32
            elif params['dtype_str'] == 'bf16': dtype = torch.bfloat16
        element_size = torch.tensor([], dtype=dtype).element_size()
    except KeyError:
        element_size = 2 # Default to 2 bytes (fp16)

    bytes_read_x = M * N * element_size
    bytes_read_w = N * element_size
    bytes_read_b = N * element_size
    bytes_write_y = M * N * element_size
    bytes_write_mean = M * 4 # Mean/rstd often float32
    bytes_write_rstd = M * 4

    total_bytes = bytes_read_x + bytes_read_w + bytes_read_b + \
                  bytes_write_y + bytes_write_mean + bytes_write_rstd
    gbps = total_bytes / (ms / 1000) / 1e9
    return gbps

def calculate_layernorm_tflops(params: dict, ms: float) -> float:
    M = params['M']
    N = params['N']
    # FLOPs for LayerNorm forward:
    # 1. Mean: N additions + 1 division per row  (N ops)
    # 2. Variance: N subtractions, N squares, N additions, 1 division per row (3N+1 ops, approx 3N)
    # 3. rsqrt: (approx ~5-10 ops, let's say 5)
    # 4. Normalize: N subtractions, N multiplications per row (2N ops)
    # 5. Scale/shift: N multiplications, N additions per row (2N ops)
    # Total per row: N + 3N + 5 + 2N + 2N = 8N + 5 ops
    # Total FLOPs: M * (8*N + 5)
    flops = M * (8 * N + 5) # Simplified, actual can vary by rsqrt implementation etc.
    tflops = flops / (ms / 1000) / 1e12
    return tflops

# --- Name for the benchmark output JSON file ---
OP_NAME_FOR_BENCHMARK = "layernorm_triton_fwd_perf"

# --- Pytest parametrize for performance testing ---
# Using the same parameters as the original test_layernorm
LAYERNORM_SHAPES_FOR_PERF = [
    (1823, 781), (2048, 2048), (8192, 8192), # Some medium to large
    (4096, 10240), # LLM typical
    (1, 131072), (1, 89999), # Long sequence, batch 1
    (128, 2048), (512, 4096)
]
# Dtypes to test for performance
DTYPES_FOR_PERF = ['fp16', 'bf16', 'fp32']


@pytest.mark.parametrize('M, N', LAYERNORM_SHAPES_FOR_PERF)
@pytest.mark.parametrize('dtype_str', DTYPES_FOR_PERF)
def test_performance(M, N, dtype_str, request):
    set_seed()
    eps = 1e-5

    # Determine torch dtype
    if dtype_str == 'fp32': current_dtype = torch.float32
    elif dtype_str == 'bf16': current_dtype = torch.bfloat16
    else: current_dtype = torch.float16 # Default to fp16

    # Input tensors
    x = torch.randn(M, N, device='cuda', dtype=current_dtype)
    normalized_shape_arg = (N,) # For torch.nn.functional.layer_norm and our LayerNorm.apply
    w = torch.rand(N, device='cuda', dtype=current_dtype)
    b = torch.rand(N, device='cuda', dtype=current_dtype)

    # --- Create op_lambda for benchmarking the forward pass ---
    op_lambda = lambda: layernorm(x, normalized_shape_arg, w, b, eps)

    # --- Benchmarking ---
    # Autotuned kernels might benefit from fewer reps if tuning takes time,
    # but do_bench needs enough reps for stable measurement AFTER tuning.
    bench_config = do_bench_config(warm_up=25, repetition=100)
    benchmarker = PytestBenchmarker(op_callable=op_lambda,
                                    op_name=OP_NAME_FOR_BENCHMARK,
                                    config=bench_config)

    current_params_for_logs_and_calc = {
        "M": M, "N": N, "eps": eps, "dtype_str": dtype_str
    }

    benchmarker.run_benchmark(current_params_dict=current_params_for_logs_and_calc,
                              gbps_calculator=calculate_layernorm_gbps,
                              tflops_calculator=calculate_layernorm_tflops)
    
######################################## HELPERS for Eval ########################################     
# --- Pytest hook to save the dictionary at the end of the session ---  
def test_save_results():  
    """  
    Called after whole test run finished, right before returning the exit status to the system.  
    """
    print('Inside session finish...')
    if "_CALL_SUCCESS_" not in result_gold:
        result_gold['_CALL_SUCCESS_'] = torch.tensor([[0.0]])
    OUTPUT_FILENAME = __file__.replace('.','_') + '.pt'
    print(f"\nSaving all y_triton results to {OUTPUT_FILENAME}...")  
    # Ensure the directory for the output file exists if it's in a subdirectory  
    output_dir = os.path.dirname(OUTPUT_FILENAME)  
    if output_dir and not os.path.exists(output_dir):  
        os.makedirs(output_dir, exist_ok=True)  
    torch.save(result_gold, OUTPUT_FILENAME)       
    print(f"Successfully saved {len(result_gold)} y_triton tensors to {OUTPUT_FILENAME}.")  

def test_save_performance_results():
    """
    Called after the test_performance function finishes.
    This is a separate hook to ensure performance results are saved.
    """
    print('\nPytest session finishing... Saving benchmark results...')

    output_directory = os.path.join(os.path.dirname(__file__), "perf")  # Save in a "perf" subdirectory next to the test file
    os.makedirs(output_directory, exist_ok=True)
    
    save_all_benchmark_results(output_directory)
    print(f"All benchmark results attempted to save to: {output_directory}")
######################################## HELPERS for Eval ######################################## 


# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

# Triton kernel to multiply each element in the source tensor by a constant exponent compensator
@triton.jit
def mul_kernel(src, dst, BLOCK_SIZE: tl.constexpr):
    # Define a constant exponent compensator
    exponent_compensator: tl.constexpr = 2.0 ** (127 - 15)
    # Calculate the indices for the current program ID
    idxs = tl.program_id(0) * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    # Load elements from the source tensor
    x = tl.load(src + idxs)
    # Multiply each element by the exponent compensator
    y = x * exponent_compensator
    # Store the result in the destination tensor
    tl.store(dst + idxs, y)

# Function to launch the Triton kernel
def launch_mul_kernel(src, BLOCK_SIZE=1):
    # Create an empty tensor for the result
    dst = torch.empty(src.shape, dtype=torch.float32, device='cuda')
    # Launch the Triton kernel
    mul_kernel[(src.shape[0] // BLOCK_SIZE,)](src, dst, BLOCK_SIZE)
    return dst




##################################################################################################################################################


def test_mul():
    src = torch.tensor([8323072], dtype=torch.int32, device='cuda').view(torch.float32)
    
    test_cases = {}
    
    # Test case 1
    dst_triton_1 = launch_mul_kernel(src, BLOCK_SIZE=1)
    test_cases['test_case_1'] = dst_triton_1

    # Test case 2
    dst_triton_2 = launch_mul_kernel(src, BLOCK_SIZE=2)
    test_cases['test_case_2'] = dst_triton_2
    
    # Test case 3
    dst_triton_3 = launch_mul_kernel(src, BLOCK_SIZE=4)
    test_cases['test_case_3'] = dst_triton_3
    
    # Test case 4
    dst_triton_4 = launch_mul_kernel(src, BLOCK_SIZE=8)
    test_cases['test_case_4'] = dst_triton_4

    return test_cases

result_gold = test_mul()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.multinomial_sampling import multinomial_sampling
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('multinomial_sampling', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(4, 20):  # Adjust the range as needed for your tests
            batch_size = 2 ** i
            num_tokens = 2 ** 10  # Example token size, adjust as needed
            scores = torch.rand((batch_size, num_tokens), dtype=torch.float32)
            seeds = torch.randint(0, 2**31, (batch_size,), dtype=torch.int64)
            offsets = torch.zeros(batch_size, dtype=torch.int64)  # Example offsets
            indices = torch.arange(num_tokens).expand(batch_size, num_tokens)
            self.input_tensors.append((scores, seeds, offsets, indices))

    def to_cuda(self, input_tensor):
        scores, seeds, offsets, indices = input_tensor
        return (scores.cuda(), seeds.cuda(), offsets.cuda(), indices.cuda())

    def call_op(self, input_tensor):
        scores, seeds, offsets, indices = input_tensor
        return multinomial_sampling(scores, seeds, offsets, indices)

    def get_gbps(self, input_tensor, runtime):
        scores, _, _, _ = input_tensor
        total_bytes = scores.numel() * scores.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        scores, _, _, _ = input_tensor
        FLOPS = 2 * scores.numel()  # Example calculation, adjust based on actual operations
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def add_kernel(x_ptr,  # *Pointer* to first input vector.
               y_ptr,  # *Pointer* to second input vector.
               output_ptr,  # *Pointer* to output vector.
               n_elements,  # Size of the vector.
               BLOCK_SIZE: tl.constexpr,  # Number of elements each program should process.
               ):
    pid = tl.program_id(axis=0)  # We use a 1D launch grid so axis is 0.
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(x_ptr + offsets, mask=mask)
    y = tl.load(y_ptr + offsets, mask=mask)
    output = x + y
    tl.store(output_ptr + offsets, output, mask=mask)

def add(x: torch.Tensor, y: torch.Tensor):
    output = torch.empty_like(x)
    assert x.is_cuda and y.is_cuda and output.is_cuda
    n_elements = output.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta['BLOCK_SIZE']), )
    add_kernel[grid](x, y, output, n_elements, BLOCK_SIZE=1024)
    return output



##################################################################################################################################################


def test_add():
    torch.manual_seed(0)
    size = 98432
    x = torch.rand(size, device='cuda')
    y = torch.rand(size, device='cuda')
    
    # Test case 1
    output_triton_1 = add(x, y)
    
    # Test case 2
    size_2 = 1024
    x_2 = torch.rand(size_2, device='cuda')
    y_2 = torch.rand(size_2, device='cuda')
    output_triton_2 = add(x_2, y_2)
    
    # Test case 3
    size_3 = 2048
    x_3 = torch.rand(size_3, device='cuda')
    y_3 = torch.rand(size_3, device='cuda')
    output_triton_3 = add(x_3, y_3)
    
    # Test case 4
    size_4 = 4096
    x_4 = torch.rand(size_4, device='cuda')
    y_4 = torch.rand(size_4, device='cuda')
    output_triton_4 = add(x_4, y_4)
    
    results = {
        "test_case_1": output_triton_1,
        "test_case_2": output_triton_2,
        "test_case_3": output_triton_3,
        "test_case_4": output_triton_4
    }
    
    return results

result_gold = test_add()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl

@triton.jit
def batched_vecmat_kernel(
        A,  # shape: [dim_m, dim_k]
        B,  # shape: [dim_m, dim_n, dim_k]
        dim_m, dim_n, dim_k,
        output,
        block_m: tl.constexpr, block_n: tl.constexpr, block_k: tl.constexpr):
    m_index = tl.program_id(0)
    n_index = tl.program_id(1)
    output_tile = (m_index * block_m + tl.arange(0, block_m))[:, None] * dim_n \
        + (n_index * block_n + tl.arange(0, block_n))[None, :]

    vecmat = tl.zeros([block_m, block_n], dtype=A.dtype.element_ty)
    k_blocks = dim_k // block_k
    for k_index in range(k_blocks):
        a_tile = (m_index * block_m + tl.arange(0, block_m))[:, None] * dim_k \
            + (k_index * block_k + tl.arange(0, block_k))[None, :]
        a = tl.load(A + a_tile)

        b_tile = (m_index * block_m + tl.arange(0, block_m))[None, :, None] * dim_n * dim_k \
            + (n_index * block_n + tl.arange(0, block_n))[:, None, None] * dim_k \
            + (k_index * block_k + tl.arange(0, block_k))[None, None, :]
        b = tl.load(B + b_tile)

        expanded_a, _ = tl.broadcast(a, b)
        vecmat += tl.trans(tl.sum(expanded_a * b, axis=2))

    tl.store(output + output_tile, vecmat)


def batched_vecmat(
    M, N, K, block_m, block_n, block_k, num_warps=4, num_stages=1
):

    A = torch.randn(M, K, device='cuda', dtype=torch.float32)  # shape: [M, K]
    B = torch.randn(M, N, K, device='cuda', dtype=torch.float32)  # shape: [M, N, K]
    output = torch.zeros(M, N, device='cuda', dtype=torch.float32)  # 输出张量，shape: [M, N]

    assert K % block_k == 0, ""
    assert M % block_m == 0, ""
    assert N % block_n == 0, ""

    grid = (M // block_m, N // block_n)

    # 调用 Triton Kernel
    batched_vecmat_kernel[grid](
        A,
        B,
        M, N, K,
        output,
        block_m=block_m,
        block_n=block_n,
        block_k=block_k,
        num_warps=num_warps,
        num_stages=num_stages
    )

    return output



##################################################################################################################################################


# Function 3: Test the correctness of the Triton kernel against the reference implementation
def test_vecmat():
    M, N, K = 128, 128, 128
    block_m, block_n, block_k = 16, 32, 64

    results = {}
    output = batched_vecmat(M, N, K, block_m, block_n, block_k)
    results['test_case_1'] = output.clone()  # Store first result

    output2 = batched_vecmat(M, N, K, block_m, block_n, block_k, num_warps=2, num_stages=2)
    results['test_case_2'] = output2.clone()  # Store second result with different key

    return results

# Run the test
result_gold = test_vecmat()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch
import triton
import triton.language as tl
import functools

def is_hip() -> bool:
    return torch.version.hip is not None

def calculate_settings(n):
    MAX_FUSED_SIZE = 65536
    BLOCK_SIZE = triton.next_power_of_2(n)
    if BLOCK_SIZE > MAX_FUSED_SIZE:
        raise RuntimeError(
            f"Cannot launch Triton kernel since n = {n} exceeds "
            f"the recommended Triton blocksize = {MAX_FUSED_SIZE}."
        )

    num_warps = 4
    if BLOCK_SIZE >= 32768:
        num_warps = 32 if not is_hip() else 16
    elif BLOCK_SIZE >= 8192:
        num_warps = 16
    elif BLOCK_SIZE >= 2048:
        num_warps = 8
    return BLOCK_SIZE, num_warps

@triton.jit
def silu(x):
    return x * tl.sigmoid(x)

@triton.jit
def _swiglu_forward_kernel(
    a_ptr, b_ptr, c_ptr, stride, n_cols: tl.constexpr, BLOCK_SIZE: tl.constexpr
):
    program_id = tl.program_id(0).to(tl.int64)

    # locate start index
    a_ptr += program_id * stride
    b_ptr += program_id * stride
    c_ptr += program_id * stride

    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    # sigmoid requires type float32
    a_row = tl.load(a_ptr + col_offsets, mask=mask, other=0).to(tl.float32)
    b_row = tl.load(b_ptr + col_offsets, mask=mask, other=0)
    c_row = silu(a_row) * b_row
    tl.store(c_ptr + col_offsets, c_row, mask=mask)

@triton.jit
def _swiglu_backward_kernel(
    dc_ptr, a_ptr, b_ptr, stride, n_cols: tl.constexpr, BLOCK_SIZE: tl.constexpr
):
    program_id = tl.program_id(0).to(tl.int64)

    # locate start index
    dc_ptr += program_id * stride
    a_ptr += program_id * stride
    b_ptr += program_id * stride

    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    dc_row = tl.load(dc_ptr + col_offsets, mask=mask, other=0)
    # sigmoid requires type float32
    a_row = tl.load(a_ptr + col_offsets, mask=mask, other=0).to(tl.float32)
    b_row = tl.load(b_ptr + col_offsets, mask=mask, other=0)

    # recomputation to save memory
    sig_a = tl.sigmoid(a_row)
    silu_a = a_row * sig_a
    db_row = dc_row * silu_a
    da_row = dc_row * (silu_a * (1 - sig_a) + sig_a) * b_row

    tl.store(a_ptr + col_offsets, da_row, mask=mask)
    tl.store(b_ptr + col_offsets, db_row, mask=mask)

def swiglu_forward(a, b):
    ori_shape = a.shape

    n_cols = ori_shape[-1]
    a = a.view(-1, n_cols)
    b = b.view(-1, n_cols)
    c = torch.empty_like(a)
    n_rows = a.shape[0]

    BLOCK_SIZE, num_warps = calculate_settings(n_cols)

    _swiglu_forward_kernel[(n_rows,)](
        a,
        b,
        c,
        c.stride(-2),
        n_cols=n_cols,
        BLOCK_SIZE=BLOCK_SIZE,
        num_warps=num_warps,
    )
    return a, b, c.view(*ori_shape)

def swiglu_backward(a, b, dc):
    ori_shape = dc.shape
    n_cols = ori_shape[-1]
    dc = dc.view(-1, n_cols)
    n_rows = dc.shape[0]

    BLOCK_SIZE, num_warps = calculate_settings(n_cols)

    _swiglu_backward_kernel[(n_rows,)](
        dc,
        a,
        b,
        dc.stride(-2),
        n_cols=n_cols,
        BLOCK_SIZE=BLOCK_SIZE,
        num_warps=num_warps,
    )
    return a.view(*ori_shape), b.view(*ori_shape)




##################################################################################################################################################


import torch

# Test the swiglu_forward function
def test_swiglu():
    # Create input tensors
    a = torch.randn(4, 8, dtype=torch.float32, device='cuda')  # Example shape (4, 8)
    b = torch.randn(4, 8, dtype=torch.float32, device='cuda')  # Same shape as a
    dc = torch.randn(4, 8, dtype=torch.float32, device='cuda')  # Gradient tensor

    # Call the forward function
    a_out, b_out, c_out = swiglu_forward(a, b)

    # Call the backward function
    da_out, db_out = swiglu_backward(a, b, dc)

    # Store results in a dictionary
    results = {
        "test_case_1": (a_out, b_out, c_out, da_out, db_out)
    }

    return results

# Run the tests
result_gold = test_swiglu()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from TritonBench_v1.rotary_transform import apply_rotary
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('rotary_transform', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(2, 16):  # Choose a reasonable range for testing
            batch_size = 2 ** i
            seqlen = 128  # Fixed sequence length
            nheads = 8  # Number of attention heads
            headdim = 64  # Dimension of each head
            rotary_dim = 32  # Rotary dimension
            x = torch.rand(batch_size, seqlen, nheads, headdim, dtype=torch.float32)
            cos = torch.rand(seqlen, rotary_dim // 2, dtype=torch.float32)
            sin = torch.rand(seqlen, rotary_dim // 2, dtype=torch.float32)
            self.input_tensors.append((x, cos, sin))

    def to_cuda(self, input_tensor):
        x, cos, sin = input_tensor
        return (x.cuda(), cos.cuda(), sin.cuda())

    def call_op(self, input_tensor):
        x, cos, sin = input_tensor
        return apply_rotary(x, cos, sin)

    def get_gbps(self, input_tensor, runtime):
        x, cos, sin = input_tensor
        total_bytes = x.numel() * x.element_size() + cos.numel() * cos.element_size() + sin.numel() * sin.element_size()
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        x, cos, sin = input_tensor
        # Assuming each element in x is involved in a few operations (e.g., multiply and add)
        FLOPS = 2 * x.numel()  # Simplified estimation
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

# Correctly import the operator
from TritonBench_v1.layer_norm_triton import layer_norm
from performance_utils import Performance_Metrics, do_bench_config

import torch
import triton
import triton.language as tl

class performance_metrics(Performance_Metrics):
    def __init__(self, dtype=None, is_backward=False, **kwargs):
        super().__init__('layer_norm_triton', dtype=dtype, is_backward=is_backward, **kwargs)
        
    def get_input_tensors(self):
        self.input_tensors = []
        for i in range(4, 15):
            size = 2 ** i
            normalized_shape = (size,)
            weight = torch.rand(size, dtype=torch.float32)
            bias = torch.rand(size, dtype=torch.float32)
            input_tensor = (torch.rand(size, dtype=torch.float32), normalized_shape, weight, bias, 1e-5)
            self.input_tensors.append(input_tensor)

    def to_cuda(self, input_tensor):
        x, normalized_shape, weight, bias, eps = input_tensor
        return (x.cuda(), normalized_shape, weight.cuda(), bias.cuda(), eps)

    def call_op(self, input_tensor):
        x, normalized_shape, weight, bias, eps = input_tensor
        return layer_norm(x, normalized_shape, weight, bias, eps)

    def get_gbps(self, input_tensor, runtime):
        x = input_tensor[0]
        total_bytes = 2 * x.numel() * x.element_size()  # Read and write
        GBPS = total_bytes / (runtime / 1000) / 1e9
        return GBPS
    
    def get_tflops(self, input_tensor, runtime):
        x = input_tensor[0]
        # Assuming each element involves a few FLOPs for normalization
        FLOPS = 5 * float(x.numel())  # Approximation
        TFLOPS = FLOPS / (runtime / 1000) / 1e12
        return TFLOPS

if __name__ == '__main__':
    op_perf = performance_metrics()
    op_perf.get_input_tensors()
    op_perf.get_do_bench_config(warmup=100, rep=1000)
    op_perf.run_benchmark()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import triton
import triton.language as tl
import torch

@triton.jit
def _fwd_kernel_token_att1(
    Q, K, sm_scale, B_Loc, B_Start_Loc, B_Seqlen, max_input_len,
    Att_Out,
    stride_b_loc_b, stride_b_loc_s,
    stride_qb, stride_qh, stride_qs, stride_qd,
    stride_kb, stride_kh, stride_ks, stride_kd,
    att_stride_h, att_stride_bs,
    kv_group_num,
    BLOCK_DMODEL: tl.constexpr,
    BLOCK_N: tl.constexpr
):
    cur_batch = tl.program_id(0)
    cur_head = tl.program_id(1)
    start_n = tl.program_id(2)
    
    cur_kv_head = cur_head // kv_group_num

    offs_d = tl.arange(0, BLOCK_DMODEL)
    cur_batch_seq_len = tl.load(B_Seqlen + cur_batch)
    cur_batch_in_all_start_index = tl.load(B_Start_Loc + cur_batch)

    cur_batch_start_index = max_input_len - cur_batch_seq_len
    cur_batch_end_index = max_input_len

    off_q = cur_batch * stride_qb + cur_head * stride_qh + offs_d * stride_qd

    offs_n = start_n * BLOCK_N + tl.arange(0, BLOCK_N)

    block_stard_index = start_n * BLOCK_N
    block_mask = tl.where(block_stard_index < cur_batch_seq_len, 1, 0)

    for start_mark in range(0, block_mask, 1):
        q = tl.load(Q + off_q + start_mark)
        offs_n_new = cur_batch_start_index + offs_n
        k_loc = tl.load(B_Loc + stride_b_loc_b * cur_batch + stride_b_loc_s * offs_n_new, mask=offs_n_new < cur_batch_end_index, other=0)
        off_k = k_loc[:, None] * stride_ks + cur_kv_head * stride_kh + offs_d[None, :] * stride_kd
        k = tl.load(K + off_k, mask=offs_n_new[:, None] < cur_batch_end_index, other=0.0)
        att_value = tl.sum(q[None, :] * k, 1)
        att_value *= sm_scale
        off_o = cur_head * att_stride_h + (cur_batch_in_all_start_index + offs_n) * att_stride_bs
        tl.store(Att_Out + off_o, att_value, mask=offs_n_new < cur_batch_end_index)
    return

@torch.no_grad()
def token_att_fwd(q, k, att_out, B_Loc, B_Start_Loc, B_Seqlen, max_input_len):
    BLOCK = 32
    # shape constraints
    Lq, Lk = q.shape[-1], k.shape[-1]
    assert Lq == Lk
    assert Lk in {16, 32, 64, 128}
    sm_scale = 1.0 / (Lk ** 0.5)

    batch, head_num = B_Loc.shape[0], q.shape[1]

    grid = (batch, head_num, triton.cdiv(max_input_len, BLOCK))
    kv_group_num = q.shape[1] // k.shape[1]

    num_warps = 4 if Lk <= 64 else 8
    num_warps = 2

    _fwd_kernel_token_att1[grid](
        q, k, sm_scale, B_Loc, B_Start_Loc, B_Seqlen, max_input_len,
        att_out,
        B_Loc.stride(0), B_Loc.stride(1),
        q.stride(0), q.stride(1), q.stride(2), q.stride(3),
        k.stride(0), k.stride(1), k.stride(2), k.stride(3),
        att_out.stride(1), att_out.stride(0),
        kv_group_num=kv_group_num,
        BLOCK_DMODEL=Lk,
        BLOCK_N=BLOCK,
        num_warps=num_warps,
        num_stages=1,
    )
    return




##################################################################################################################################################


import torch

def test_token_att_fwd():
    # Define the input parameters
    batch_size = 2
    head_num = 4
    max_input_len = 64
    d_model = 32  # This should be one of {16, 32, 64, 128}

    # Create random input tensors
    q = torch.randn((batch_size, head_num, max_input_len, d_model), dtype=torch.float32, device='cuda')
    k = torch.randn((batch_size, head_num, max_input_len, d_model), dtype=torch.float32, device='cuda')
    att_out = torch.zeros((batch_size, head_num, max_input_len), dtype=torch.float32, device='cuda')

    # Create B_Loc, B_Start_Loc, B_Seqlen
    B_Loc = torch.randint(0, max_input_len, (batch_size, max_input_len), dtype=torch.int32, device='cuda')
    B_Start_Loc = torch.randint(0, max_input_len, (batch_size,), dtype=torch.int32, device='cuda')
    B_Seqlen = torch.randint(1, max_input_len + 1, (batch_size,), dtype=torch.int32, device='cuda')

    # Dictionary to store results for each test case
    results = {}

    # Test case 1
    token_att_fwd(q, k, att_out, B_Loc, B_Start_Loc, B_Seqlen, max_input_len)
    results['test_case_1'] = att_out.clone()

    # Additional test cases to cover more branches
    # Test case 2: Different max_input_len
    max_input_len_2 = 32
    att_out_2 = torch.zeros((batch_size, head_num, max_input_len_2), dtype=torch.float32, device='cuda')
    token_att_fwd(q, k, att_out_2, B_Loc, B_Start_Loc, B_Seqlen, max_input_len_2)
    results['test_case_2'] = att_out_2.clone()

    # Test case 3: Different d_model
    d_model_3 = 64
    q_3 = torch.randn((batch_size, head_num, max_input_len, d_model_3), dtype=torch.float32, device='cuda')
    k_3 = torch.randn((batch_size, head_num, max_input_len, d_model_3), dtype=torch.float32, device='cuda')
    att_out_3 = torch.zeros((batch_size, head_num, max_input_len), dtype=torch.float32, device='cuda')
    token_att_fwd(q_3, k_3, att_out_3, B_Loc, B_Start_Loc, B_Seqlen, max_input_len)
    results['test_case_3'] = att_out_3.clone()

    # Test case 4: Different batch size
    batch_size_4 = 4
    q_4 = torch.randn((batch_size_4, head_num, max_input_len, d_model), dtype=torch.float32, device='cuda')
    k_4 = torch.randn((batch_size_4, head_num, max_input_len, d_model), dtype=torch.float32, device='cuda')
    att_out_4 = torch.zeros((batch_size_4, head_num, max_input_len), dtype=torch.float32, device='cuda')
    B_Loc_4 = torch.randint(0, max_input_len, (batch_size_4, max_input_len), dtype=torch.int32, device='cuda')
    B_Start_Loc_4 = torch.randint(0, max_input_len, (batch_size_4,), dtype=torch.int32, device='cuda')
    B_Seqlen_4 = torch.randint(1, max_input_len + 1, (batch_size_4,), dtype=torch.int32, device='cuda')
    token_att_fwd(q_4, k_4, att_out_4, B_Loc_4, B_Start_Loc_4, B_Seqlen_4, max_input_len)
    results['test_case_4'] = att_out_4.clone()

    return results

# Execute the test function
result_gold = test_token_att_fwd()



# Modifications Copyright(C)[2025] Advanced Micro Devices, Inc. All rights reserved.
# https://github.com/thunlp/TritonBench - Apache License 2.0
import torch

import triton
import triton.language as tl

@triton.jit
def _fwd_kernel_destindex_copy_quantize_kv(
    K,
    Dest_loc,
    Out,
    Out_scale,
    stride_k_bs,
    stride_k_h,
    stride_k_d,
    stride_o_bs,
    stride_o_h,
    stride_o_d,
    stride_os_bs,
    stride_os_h,
    stride_os_d,
    head_num,
    head_dim,
    BLOCK_DMODEL: tl.constexpr,
    BLOCK_HEAD: tl.constexpr,
):
    cur_index = tl.program_id(0)
    offs_h = tl.arange(0, BLOCK_HEAD)
    offs_d = tl.arange(0, BLOCK_DMODEL)

    dest_index = tl.load(Dest_loc + cur_index)
    src_data = tl.load(
        K + cur_index * stride_k_bs + offs_h[:, None] * stride_k_h + stride_k_d * offs_d[None, :],
        mask=(offs_h[:, None] < head_num) & (offs_d[None, :] < head_dim),
        other=0.0,
    )
    abs_data = tl.abs(src_data)
    data_scale = (tl.max(abs_data, axis=1) / 127.0).to(Out_scale.dtype.element_ty)[:, None]
    q_src_data = (src_data / data_scale).to(tl.int8)
    o_ptrs = Out + dest_index * stride_o_bs + stride_o_h * offs_h[:, None] + stride_o_d * offs_d[None, :]
    os_ptrs = Out_scale + dest_index * stride_os_bs + stride_os_h * offs_h[:, None]
    tl.store(o_ptrs, q_src_data, mask=(offs_h[:, None] < head_num) & (offs_d[None, :] < head_dim))
    tl.store(os_ptrs, data_scale, mask=(offs_h[:, None] < head_num))


@torch.no_grad()
def destindex_copy_quantize_kv(K, DestLoc, Out, Out_scale):
    seq_len = DestLoc.shape[0]
    head_num = K.shape[1]
    head_dim = K.shape[2]
    assert K.shape[1] == Out.shape[1] and K.shape[2] == Out.shape[2]
    BLOCK_HEAD = triton.next_power_of_2(head_num)
    BLOCK_DMODEL = triton.next_power_of_2(head_dim)
    grid = (seq_len,)
    num_warps = 1

    _fwd_kernel_destindex_copy_quantize_kv[grid](
        K,
        DestLoc,
        Out,
        Out_scale,
        K.stride(0),
        K.stride(1),
        K.stride(2),
        Out.stride(0),
        Out.stride(1),
        Out.stride(2),
        Out_scale.stride(0),
        Out_scale.stride(1),
        Out_scale.stride(2),
        head_num,
        head_dim,
        BLOCK_DMODEL=BLOCK_DMODEL,
        BLOCK_HEAD=BLOCK_HEAD,
        num_warps=num_warps,
        num_stages=1,
    )
    return




##################################################################################################################################################


def test_destindex_copy_quantize_kv():
    B, N_CTX, H, D = 32, 1024, 12, 96
    src = torch.randn((B * N_CTX, H, D), dtype=torch.float16).cuda()
    dest_loc = torch.arange(0, B * N_CTX, dtype=torch.int32).cuda()
    value_dest = torch.randn((B * N_CTX, H, D), dtype=torch.float16).cuda().to(torch.int8)
    scale_dest = torch.randn((B * N_CTX, H, 1), dtype=torch.float16).cuda()

    # Test case 1
    destindex_copy_quantize_kv(src, dest_loc, value_dest, scale_dest)
    result_1 = {
        "value_dest": value_dest.clone(),
        "scale_dest": scale_dest.clone()
    }

    # Test case 2: Different dimensions
    B, N_CTX, H, D = 16, 512, 8, 64
    src = torch.randn((B * N_CTX, H, D), dtype=torch.float16).cuda()
    dest_loc = torch.arange(0, B * N_CTX, dtype=torch.int32).cuda()
    value_dest = torch.randn((B * N_CTX, H, D), dtype=torch.float16).cuda().to(torch.int8)
    scale_dest = torch.randn((B * N_CTX, H, 1), dtype=torch.float16).cuda()
    destindex_copy_quantize_kv(src, dest_loc, value_dest, scale_dest)
    result_2 = {
        "value_dest": value_dest.clone(),
        "scale_dest": scale_dest.clone()
    }

    # Test case 3: Different data types
    src = torch.randn((B * N_CTX, H, D), dtype=torch.float32).cuda()
    destindex_copy_quantize_kv(src, dest_loc, value_dest, scale_dest)
    result_3 = {
        "value_dest": value_dest.clone(),
        "scale_dest": scale_dest.clone()
    }

    # Test case 4: Edge case with minimal dimensions
    B, N_CTX, H, D = 1, 1, 1, 1
    src = torch.randn((B * N_CTX, H, D), dtype=torch.float16).cuda()
    dest_loc = torch.arange(0, B * N_CTX, dtype=torch.int32).cuda()
    value_dest = torch.randn((B * N_CTX, H, D), dtype=torch.float16).cuda().to(torch.int8)
    scale_dest = torch.randn((B * N_CTX, H, 1), dtype=torch.float16).cuda()
    destindex_copy_quantize_kv(src, dest_loc, value_dest, scale_dest)
    result_4 = {
        "value_dest": value_dest.clone(),
        "scale_dest": scale_dest.clone()
    }

    return {
        "test_case_1": result_1,
        "test_case_2": result_2,
        "test_case_3": result_3,
        "test_case_4": result_4
    }

result_gold = test_destindex_copy_quantize_kv()



import os
import subprocess
from multiprocessing import Pool, Lock, Value
from tqdm import tqdm

script_dir = "./tmp"
log_dir = "./logs"
gpu_count = 8

scripts = sorted([f for f in os.listdir(script_dir) if f.endswith(".py")])
scripts = [os.path.join(script_dir, script) for script in scripts]
total_scripts = len(scripts)  

progress = Value('i', 0)
progress_lock = Lock()

def run_script(args):
    gpu_id, script = args

    script_name = os.path.basename(script)
    log_file = os.path.join(log_dir, f"{script_name}.log")
    err_file = os.path.join(log_dir, f"{script_name}.err")

    cmd = f"HIP_VISIBLE_DEVICES={gpu_id} python {script}"
    # print(f"Running: {cmd}")

    with open(log_file, "w") as log, open(err_file, "w") as err:
        process = subprocess.Popen(cmd, shell=True, stdout=log, stderr=err)
        process.wait()

    with progress_lock:
        progress.value += 1
        tqdm.write(f"✅ finished {progress.value}/{total_scripts}: {script_name}")

os.makedirs(log_dir, exist_ok=True)

if __name__ == "__main__":
    with Pool(processes=gpu_count) as pool, tqdm(total=total_scripts, desc="Process", ncols=80) as pbar:
        args_list = [(i % gpu_count, scripts[i]) for i in range(total_scripts)]
        
        for _ in pool.imap(run_script, args_list):
            pbar.update(1)

        pool.close()
        pool.join()